# -----------------------------------------------------------------------------

class STType(object):
    """
    ST types.

    - `SHELL`:   User-defined.
    - `SHIFT`:   Shift from origin.
    - `SINVERT`: Shift-and-invert.
    - `CAYLEY`:  Cayley transform.
    - `PRECOND`: Preconditioner.
    - `FILTER`:  Polynomial filter.
    """
    SHELL   = S_(STSHELL)
    SHIFT   = S_(STSHIFT)
    SINVERT = S_(STSINVERT)
    CAYLEY  = S_(STCAYLEY)
    PRECOND = S_(STPRECOND)
    FILTER  = S_(STFILTER)

class STMatMode(object):
    """
    ST matrix mode.

    - `COPY`:    A working copy of the matrix is created.
    - `INPLACE`: The operation is computed in-place.
    - `SHELL`:   The matrix :math:`A - \sigma B` is handled as an
                 implicit matrix.
    """
    COPY    = ST_MATMODE_COPY
    INPLACE = ST_MATMODE_INPLACE
    SHELL   = ST_MATMODE_SHELL

class STFilterType(object):
    """
    ST filter type.

    - ``FILTLAN``:  An adapted implementation of the Filtered Lanczos Package.
    - ``CHEBYSEV``: A polynomial filter based on a truncated Chebyshev series.
    """
    FILTLAN   = ST_FILTER_FILTLAN
    CHEBYSHEV = ST_FILTER_CHEBYSHEV

class STFilterDamping(object):
    """
    ST filter damping.

    - `NONE`:    No damping
    - `JACKSON`: Jackson damping
    - `LANCZOS`: Lanczos damping
    - `FEJER`:   Fejer damping
    """
    NONE    = ST_FILTER_DAMPING_NONE
    JACKSON = ST_FILTER_DAMPING_JACKSON
    LANCZOS = ST_FILTER_DAMPING_LANCZOS
    FEJER   = ST_FILTER_DAMPING_FEJER

# -----------------------------------------------------------------------------

cdef class ST(Object):

    """ST."""

    Type          = STType
    MatMode       = STMatMode
    FilterType    = STFilterType
    FilterDamping = STFilterDamping

    def __cinit__(self):
        self.obj = <PetscObject*> &self.st
        self.st = NULL

    def view(self, Viewer viewer=None) -> None:
        """
        Print the ST data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.
        """
        cdef PetscViewer vwr = def_Viewer(viewer)
        CHKERR( STView(self.st, vwr) )

    def destroy(self) -> Self:
        """
        Destroy the ST object.

        Collective.
        """
        CHKERR( STDestroy(&self.st) )
        self.st = NULL
        return self

    def reset(self) -> None:
        """
        Reset the ST object.

        Collective.
        """
        CHKERR( STReset(self.st) )

    def create(self, comm: Comm | None = None) -> Self:
        """
        Create the ST object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all processes.
        """
        cdef MPI_Comm ccomm = def_Comm(comm, SLEPC_COMM_DEFAULT())
        cdef SlepcST newst = NULL
        CHKERR( STCreate(ccomm, &newst) )
        CHKERR( SlepcCLEAR(self.obj) ); self.st = newst
        return self

    def setType(self, st_type: Type | str) -> None:
        """
        Set the particular spectral transformation to be used.

        Logically collective.

        Parameters
        ----------
        st_type
            The spectral transformation to be used.

        Notes
        -----
        See `ST.Type` for available methods. The default is
        `ST.Type.SHIFT` with a zero shift.  Normally, it is best to
        use `setFromOptions()` and then set the ST type from the
        options database rather than by using this routine.  Using the
        options database provides the user with maximum flexibility in
        evaluating the different available methods.
        """
        cdef SlepcSTType cval = NULL
        st_type = str2bytes(st_type, &cval)
        CHKERR( STSetType(self.st, cval) )

    def getType(self) -> str:
        """
        Get the ST type of this object.

        Not collective.

        Returns
        -------
        str
            The spectral transformation currently being used.
        """
        cdef SlepcSTType st_type = NULL
        CHKERR( STGetType(self.st, &st_type) )
        return bytes2str(st_type)

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for searching for all ST options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all ST option requests.

        Notes
        -----
        A hyphen (``-``) must NOT be given at the beginning of the
        prefix name.  The first character of all runtime options is
        AUTOMATICALLY the hyphen.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( STSetOptionsPrefix(self.st, cval) )

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for searching for all ST options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this ST object.
        """
        cdef const char *prefix = NULL
        CHKERR( STGetOptionsPrefix(self.st, &prefix) )
        return bytes2str(prefix)

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for searching for all ST options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all ST option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( STAppendOptionsPrefix(self.st, cval) )

    def setFromOptions(self) -> None:
        """
        Set ST options from the options database.

        Collective.

        This routine must be called before `setUp()` if the user is to be
        allowed to set the solver type.

        Notes
        -----
        To see all options, run your program with the -help option.
        """
        CHKERR( STSetFromOptions(self.st) )

    #

    def setShift(self, shift: Scalar) -> None:
        """
        Set the shift associated with the spectral transformation.

        Collective.

        Parameters
        ----------
        shift
            The value of the shift.

        Notes
        -----
        In some spectral transformations, changing the shift may have
        associated a lot of work, for example recomputing a
        factorization.
        """
        cdef PetscScalar sval = asScalar(shift)
        CHKERR( STSetShift(self.st, sval) )

    def getShift(self) -> Scalar:
        """
        Get the shift associated with the spectral transformation.

        Not collective.

        Returns
        -------
        Scalar
            The value of the shift.
        """
        cdef PetscScalar sval = 0
        CHKERR( STGetShift(self.st, &sval) )
        return toScalar(sval)

    def setTransform(self, flag: bool = True) -> None:
        """
        Set a flag to indicate whether the transformed matrices are computed or not.

        Logically collective.

        Parameters
        ----------
        flag
            This flag is intended for the case of polynomial
            eigenproblems solved via linearization.
            If this flag is False (default) the spectral transformation
            is applied to the linearization (handled by the eigensolver),
            otherwise it is applied to the original problem.
        """
        cdef PetscBool sval = asBool(flag)
        CHKERR( STSetTransform(self.st, sval) )

    def getTransform(self) -> bool:
        """
        Get the flag indicating whether the transformed matrices are computed or not.

        Not collective.

        Returns
        -------
        bool
            This flag is intended for the case of polynomial
            eigenproblems solved via linearization.
            If this flag is False (default) the spectral transformation
            is applied to the linearization (handled by the eigensolver),
            otherwise it is applied to the original problem.
        """
        cdef PetscBool sval = PETSC_FALSE
        CHKERR( STGetTransform(self.st, &sval) )
        return toBool(sval)

    def setMatMode(self, mode: MatMode) -> None:
        """
        Set a flag to indicate how the matrix is being shifted.

        Logically collective.

        Set a flag to indicate how the matrix is being shifted in the
        shift-and-invert and Cayley spectral transformations.

        Parameters
        ----------
        mode
            The mode flag.

        Notes
        -----
        By default (`ST.MatMode.COPY`), a copy of matrix :math:`A` is made
        and then this copy is shifted explicitly, e.g.
        :math:`A \leftarrow (A - s B)`.

        With `ST.MatMode.INPLACE`, the original matrix :math:`A` is shifted at
        `setUp()` and unshifted at the end of the computations. With respect to
        the previous one, this mode avoids a copy of matrix :math:`A`. However,
        a backdraw is that the recovered matrix might be slightly different
        from the original one (due to roundoff).

        With `ST.MatMode.SHELL`, the solver works with an implicit shell matrix
        that represents the shifted matrix. This mode is the most efficient in
        creating the shifted matrix but it places serious limitations to the
        linear solves performed in each iteration of the eigensolver
        (typically, only iterative solvers with Jacobi preconditioning can be
        used).

        In the case of generalized problems, in the two first modes the matrix
        :math:`A - s B` has to be computed explicitly. The efficiency of
        this computation can be controlled with `setMatStructure()`.
        """
        cdef SlepcSTMatMode val = mode
        CHKERR( STSetMatMode(self.st, val) )

    def getMatMode(self) -> MatMode:
        """
        Get a flag that indicates how the matrix is being shifted.

        Not collective.

        Get a flag that indicates how the matrix is being shifted in
        the shift-and-invert and Cayley spectral transformations.

        Returns
        -------
        MatMode
            The mode flag.
        """
        cdef SlepcSTMatMode val = ST_MATMODE_INPLACE
        CHKERR( STGetMatMode(self.st, &val) )
        return val

    def setMatrices(self, operators: list[Mat]) -> None:
        """
        Set the matrices associated with the eigenvalue problem.

        Collective.

        Parameters
        ----------
        operators
            The matrices associated with the eigensystem.
        """
        operators = tuple(operators)
        cdef PetscMat *mats = NULL
        cdef Py_ssize_t k=0, n = len(operators)
        cdef tmp = allocate(<size_t>n*sizeof(PetscMat),<void**>&mats)
        for k from 0 <= k < n: mats[k] = (<Mat?>operators[k]).mat
        CHKERR( STSetMatrices(self.st, <PetscInt>n, mats) )

    def getMatrices(self) -> list[petsc4py.PETSc.Mat]:
        """
        Get the matrices associated with the eigenvalue problem.

        Collective.

        Returns
        -------
        list of petsc4py.PETSc.Mat
            The matrices associated with the eigensystem.
        """
        cdef Mat A
        cdef PetscMat mat = NULL
        cdef PetscInt k=0, n=0
        CHKERR( STGetNumMatrices(self.st, &n) )
        cdef object operators = []
        for k from 0 <= k < n:
            CHKERR( STGetMatrix(self.st, k, &mat) )
            A = Mat(); A.mat = mat; CHKERR( PetscINCREF(A.obj) )
            operators.append(A)
        return tuple(operators)

    def setMatStructure(self, structure: petsc4py.PETSc.Mat.Structure) -> None:
        """
        Set an internal Mat.Structure attribute.

        Logically collective.

        Set an internal Mat.Structure attribute to indicate which is the
        relation of the sparsity pattern of the two matrices :math:`A` and
        :math:`B` constituting the generalized eigenvalue problem. This
        function has no effect in the case of standard eigenproblems.

        Parameters
        ----------
        structure
            Either same, different, or a subset of the non-zero
            sparsity pattern.

        Notes
        -----
        By default, the sparsity patterns are assumed to be
        different. If the patterns are equal or a subset then it is
        recommended to set this attribute for efficiency reasons (in
        particular, for internal *AXPY()* matrix operations).
        """
        cdef PetscMatStructure val = matstructure(structure)
        CHKERR( STSetMatStructure(self.st, val) )

    def getMatStructure(self) -> petsc4py.PETSc.Mat.Structure:
        """
        Get the internal Mat.Structure attribute.

        Not collective.

        Get the internal Mat.Structure attribute to indicate which is
        the relation of the sparsity pattern of the matrices.

        Returns
        -------
        petsc4py.PETSc.Mat.Structure
            The structure flag.
        """
        cdef PetscMatStructure val
        CHKERR( STGetMatStructure(self.st, &val) )
        return val

    def setKSP(self, KSP ksp: petsc4py.PETSc.KSP) -> None:
        """
        Set the ``KSP`` object associated with the spectral transformation.

        Collective.

        Parameters
        ----------
        `petsc4py.PETSc.KSP`
            The linear solver object.
        """
        CHKERR( STSetKSP(self.st, ksp.ksp) )

    def getKSP(self) -> KSP:
        """
        Get the ``KSP`` object associated with the spectral transformation.

        Collective.

        Returns
        -------
        `petsc4py.PETSc.KSP`
            The linear solver object.

        Notes
        -----
        On output, the internal value of `petsc4py.PETSc.KSP` can be ``NULL`` if the
        combination of eigenproblem type and selected transformation
        does not require to solve a linear system of equations.
        """
        cdef KSP ksp = KSP()
        CHKERR( STGetKSP(self.st, &ksp.ksp) )
        CHKERR( PetscINCREF(ksp.obj) )
        return ksp

    def setPreconditionerMat(self, Mat P = None) -> None:
        """
        Set the matrix to be used to build the preconditioner.

        Collective.

        Parameters
        ----------
        P
            The matrix that will be used in constructing the preconditioner.
        """
        cdef PetscMat Pmat = P.mat if P is not None else <PetscMat>NULL
        CHKERR( STSetPreconditionerMat(self.st, Pmat) )

    def getPreconditionerMat(self) -> petsc4py.PETSc.Mat:
        """
        Get the matrix previously set by setPreconditionerMat().

        Not collective.

        Returns
        -------
        petsc4py.PETSc.Mat
            The matrix that will be used in constructing the preconditioner.
        """
        cdef Mat P = Mat()
        CHKERR( STGetPreconditionerMat(self.st, &P.mat) )
        CHKERR( PetscINCREF(P.obj) )
        return P

    def setSplitPreconditioner(self, operators: list[petsc4py.PETSc.Mat], structure: petsc4py.PETSc.Mat.Structure | None = None) -> None:
        """
        Set the matrices to be used to build the preconditioner.

        Collective.

        Parameters
        ----------
        operators
            The matrices associated with the preconditioner.
        """
        operators = tuple(operators)
        cdef PetscMatStructure cstructure = matstructure(structure)
        cdef PetscMat *mats = NULL
        cdef Py_ssize_t k=0, n = len(operators)
        cdef tmp = allocate(<size_t>n*sizeof(PetscMat),<void**>&mats)
        for k from 0 <= k < n: mats[k] = (<Mat?>operators[k]).mat
        CHKERR( STSetSplitPreconditioner(self.st, <PetscInt>n, mats, cstructure) )

    def getSplitPreconditioner(self) -> tuple[list[petsc4py.PETSc.Mat], petsc4py.PETSc.Mat.Structure]:
        """
        Get the matrices to be used to build the preconditioner.

        Not collective.

        Returns
        -------
        list of petsc4py.PETSc.Mat
            The list of matrices associated with the preconditioner.
        petsc4py.PETSc.Mat.Structure
            The structure flag.
        """
        cdef PetscInt k=0,n=0
        cdef PetscMatStructure cstructure
        cdef PetscMat mat = NULL
        CHKERR( STGetSplitPreconditionerInfo(self.st, &n, &cstructure) )
        cdef object operators = []
        for k from 0 <= k < n:
            CHKERR( STGetSplitPreconditionerTerm(self.st, k, &mat) )
            A = Mat(); A.mat = mat; CHKERR( PetscINCREF(A.obj) )
            operators.append(A)
        return tuple(operators, cstructure)

    #

    def setUp(self) -> None:
        """
        Prepare for the use of a spectral transformation.

        Collective.
        """
        CHKERR( STSetUp(self.st) )

    def apply(self, Vec x, Vec y) -> None:
        """
        Apply the spectral transformation operator to a vector.

        Collective.

        Apply the spectral transformation operator to a vector, for instance
        :math:`(A - s B)^{-1} B` in the case of the shift-and-invert
        transformation and generalized eigenproblem.

        Parameters
        ----------
        x
            The input vector.
        y
            The result vector.
        """
        CHKERR( STApply(self.st, x.vec, y.vec) )

    def applyTranspose(self, Vec x, Vec y) -> None:
        """
        Apply the transpose of the operator to a vector.

        Collective.

        Apply the transpose of the operator to a vector, for instance
        :math:`B^T(A - s B)^{-T}` in the case of the shift-and-invert
        transformation and generalized eigenproblem.

        Parameters
        ----------
        x
            The input vector.
        y
            The result vector.
        """
        CHKERR( STApplyTranspose(self.st, x.vec, y.vec) )

    def applyHermitianTranspose(self, Vec x, Vec y) -> None:
        """
        Apply the hermitian-transpose of the operator to a vector.

        Collective.

        Apply the hermitian-transpose of the operator to a vector, for instance
        :math:`B^H(A - s B)^{-H}` in the case of the shift-and-invert
        transformation and generalized eigenproblem.

        Parameters
        ----------
        x
            The input vector.
        y
            The result vector.
        """
        CHKERR( STApplyHermitianTranspose(self.st, x.vec, y.vec) )

    def applyMat(self, Mat x, Mat y) -> None:
        """
        Apply the spectral transformation operator to a matrix.

        Collective.

        Apply the spectral transformation operator to a matrix, for instance
        :math:`(A - s B)^{-1} B` in the case of the shift-and-invert
        transformation and generalized eigenproblem.

        Parameters
        ----------
        x
            The input matrix.
        y
            The result matrix.
        """
        CHKERR( STApplyMat(self.st, x.mat, y.mat) )

    def getOperator(self) -> petsc4py.PETSc.Mat:
        """
        Get a shell matrix that represents the operator of the spectral transformation.

        Collective.

        Returns
        -------
        petsc4py.PETSc.Mat
            Operator matrix.
        """
        cdef Mat op = Mat()
        CHKERR( STGetOperator(self.st, &op.mat) )
        CHKERR( PetscINCREF(op.obj) )
        return op

    def restoreOperator(self, Mat op) -> None:
        """
        Restore the previously seized operator matrix.

        Logically collective.

        Parameters
        ----------
        op
            Operator matrix previously obtained with getOperator().
        """
        CHKERR( PetscObjectDereference(<PetscObject>op.mat) )
        CHKERR( STRestoreOperator(self.st, &op.mat) )

    #

    def setCayleyAntishift(self, tau: Scalar) -> None:
        """
        Set the value of the anti-shift for the Cayley spectral transformation.

        Logically collective.

        Parameters
        ----------
        tau
            The anti-shift.

        Notes
        -----
        In the generalized Cayley transform, the operator can be expressed as
        :math:`OP = inv(A - \sigma B) (A + tau B)`. This function sets
        the value of :math:`tau`.  Use `setShift()` for setting
        :math:`\sigma`.
        """
        cdef PetscScalar sval = asScalar(tau)
        CHKERR( STCayleySetAntishift(self.st, sval) )

    def getCayleyAntishift(self) -> Scalar:
        """
        Get the value of the anti-shift for the Cayley spectral transformation.

        Not collective.

        Returns
        -------
        Scalar
            The anti-shift.
        """
        cdef PetscScalar sval = 0
        CHKERR( STCayleyGetAntishift(self.st, &sval) )
        return toScalar(sval)

    def setFilterType(self, filter_type: FilterType) -> None:
        """
        Set the method to be used to build the polynomial filter.

        Logically collective.

        Parameter
        ---------
        filter_type
            The type of filter.
        """
        cdef SlepcSTFilterType val = filter_type
        CHKERR( STFilterSetType(self.st, val) )

    def getFilterType(self) -> FilterType:
        """
        Get the method to be used to build the polynomial filter.

        Not collective.

        Returns
        -------
        FilterType
            The type of filter.
        """
        cdef SlepcSTFilterType val = ST_FILTER_FILTLAN
        CHKERR( STFilterGetType(self.st, &val) )
        return val

    def setFilterInterval(self, inta: float, intb: float) -> None:
        """
        Set the interval containing the desired eigenvalues.

        Logically collective.

        Parameters
        ----------
        inta
            The left end of the interval.
        intb
            The right end of the interval.

        Notes
        -----
        The filter will be configured to emphasize eigenvalues contained
        in the given interval, and damp out eigenvalues outside it. If the
        interval is open, then the filter is low- or high-pass, otherwise
        it is mid-pass.

        Common usage is to set the interval in `EPS` with `EPS.setInterval()`.

        The interval must be contained within the numerical range of the
        matrix, see `ST.setFilterRange()`.
        """
        cdef PetscReal rval1 = asReal(inta)
        cdef PetscReal rval2 = asReal(intb)
        CHKERR( STFilterSetInterval(self.st, rval1, rval2) )

    def getFilterInterval(self) -> tuple[float, float]:
        """
        Get the interval containing the desired eigenvalues.

        Not collective.

        Returns
        -------
        inta: float
            The left end of the interval.
        intb: float
            The right end of the interval.
        """
        cdef PetscReal inta = 0
        cdef PetscReal intb = 0
        CHKERR( STFilterGetInterval(self.st, &inta, &intb) )
        return (toReal(inta), toReal(intb))

    def setFilterRange(self, left: float, right: float) -> None:
        """
        Set the numerical range (or field of values) of the matrix.

        Logically collective.

        Set the numerical range (or field of values) of the matrix, that is,
        the interval containing all eigenvalues.

        Parameters
        ----------
        left
            The left end of the interval.
        right
            The right end of the interval.

        Notes
        -----
        The filter will be most effective if the numerical range is tight,
        that is, left and right are good approximations to the leftmost and
        rightmost eigenvalues, respectively.
        """
        cdef PetscReal rval1 = asReal(left)
        cdef PetscReal rval2 = asReal(right)
        CHKERR( STFilterSetRange(self.st, rval1, rval2) )

    def getFilterRange(self) -> tuple[float, float]:
        """
        Get the interval containing all eigenvalues.

        Not collective.

        Returns
        -------
        left: float
            The left end of the interval.
        right: float
            The right end of the interval.
        """
        cdef PetscReal left = 0
        cdef PetscReal right = 0
        CHKERR( STFilterGetRange(self.st, &left, &right) )
        return (toReal(left), toReal(right))

    def setFilterDegree(self, deg: int) -> None:
        """
        Set the degree of the filter polynomial.

        Logically collective.

        Parameters
        ----------
        deg
            The polynomial degree.
        """
        cdef PetscInt val = asInt(deg)
        CHKERR( STFilterSetDegree(self.st, val) )

    def getFilterDegree(self) -> int:
        """
        Get the degree of the filter polynomial.

        Not collective.

        Returns
        -------
        int
            The polynomial degree.
        """
        cdef PetscInt val = 0
        CHKERR( STFilterGetDegree(self.st, &val) )
        return toInt(val)

    def setFilterDamping(self, damping: FilterDamping) -> None:
        """
        Set the type of damping to be used in the polynomial filter.

        Logically collective.

        Parameter
        ---------
        damping
            The type of damping.
        """
        cdef SlepcSTFilterDamping val = damping
        CHKERR( STFilterSetDamping(self.st, val) )

    def getFilterDamping(self) -> FilterDamping:
        """
        Get the type of damping used in the polynomial filter.

        Not collective.

        Returns
        -------
        FilterDamping
            The type of damping.
        """
        cdef SlepcSTFilterDamping val = ST_FILTER_DAMPING_NONE
        CHKERR( STFilterGetDamping(self.st, &val) )
        return val

    #

    property shift:
        """Value of the shift."""
        def __get__(self) -> float:
            return self.getShift()
        def __set__(self, value):
            self.setShift(value)

    property transform:
        """If the transformed matrices are computed."""
        def __get__(self) -> bool:
            return self.getTransform()
        def __set__(self, value):
            self.setTransform(value)

    property mat_mode:
        """How the transformed matrices are being stored in the ST."""
        def __get__(self) -> STMatMode:
            return self.getMatMode()
        def __set__(self, value):
            self.setMatMode(value)

    property mat_structure:
        """Relation of the sparsity pattern of all ST matrices."""
        def __get__(self) -> MatStructure:
            return self.getMatStructure()
        def __set__(self, value):
            self.setMatStructure(value)

    property ksp:
        """KSP object associated with the spectral transformation."""
        def __get__(self) -> KSP:
            return self.getKSP()
        def __set__(self, value):
            self.setKSP(value)

# -----------------------------------------------------------------------------

del STType
del STMatMode

# -----------------------------------------------------------------------------



"""Type stubs for SLEPc DS module."""

from enum import IntEnum, StrEnum
from typing import Self, Sequence

from petsc4py.PETSc import Comm, Mat, Viewer
from petsc4py.typing import ArrayReal, ArrayScalar

class DSType(StrEnum):
    """DS type."""

    HEP = ...
    """Hermitian eigenvalue problem."""
    NHEP = ...
    """Non-Hermitian eigenvalue problem."""
    GHEP = ...
    """Generalized Hermitian eigenvalue problem."""
    GHIEP = ...
    """Generalized Hermitian-indefinite eigenvalue problem."""
    GNHEP = ...
    """Generalized non-Hermitian eigenvalue problem."""
    NHEPTS = ...
    """Non-Hermitian eigenvalue problem with two-sided approach."""
    SVD = ...
    """Singular value decomposition."""
    HSVD = ...
    """Hyperbolic singular value decomposition."""
    GSVD = ...
    """Generalized singular value decomposition."""
    PEP = ...
    """Polynomial eigenvalue problem."""
    NEP = ...
    """Nonlinear eigenvalue problem."""

class DSStateType(IntEnum):
    """DS state types.

    - `RAW`:          Not processed yet.
    - `INTERMEDIATE`: Reduced to Hessenberg or tridiagonal form (or equivalent).
    - `CONDENSED`:    Reduced to Schur or diagonal form (or equivalent).
    - `TRUNCATED`:    Condensed form truncated to a smaller size.
    """

    RAW = ...
    """Not processed yet."""
    INTERMEDIATE = ...
    """Reduced to Hessenberg or tridiagonal form (or equivalent)."""
    CONDENSED = ...
    """Reduced to Schur or diagonal form (or equivalent)."""
    TRUNCATED = ...
    """Condensed form truncated to a smaller size."""

class DSMatType(IntEnum):
    """To refer to one of the matrices stored internally in DS.

    - `A`:  first matrix of eigenproblem/singular value problem.
    - `B`:  second matrix of a generalized eigenproblem.
    - `C`:  third matrix of a quadratic eigenproblem.
    - `T`:  tridiagonal matrix.
    - `D`:  diagonal matrix.
    - `Q`:  orthogonal matrix of (right) Schur vectors.
    - `Z`:  orthogonal matrix of left Schur vectors.
    - `X`:  right eigenvectors.
    - `Y`:  left eigenvectors.
    - `U`:  left singular vectors.
    - `V`:  right singular vectors.
    - `W`:  workspace matrix.
    """

    A = ...
    """First matrix of eigenproblem/singular value problem."""
    B = ...
    """Second matrix of a generalized eigenproblem."""
    C = ...
    """Third matrix of a quadratic eigenproblem."""
    T = ...
    """Tridiagonal matrix."""
    D = ...
    """Diagonal matrix."""
    Q = ...
    """Orthogonal matrix of (right) Schur vectors."""
    Z = ...
    """Orthogonal matrix of left Schur vectors."""
    X = ...
    """Right eigenvectors."""
    Y = ...
    """Left eigenvectors."""
    U = ...
    """Left singular vectors."""
    V = ...
    """Right singular vectors."""
    W = ...
    """Workspace matrix."""

class DSParallelType(IntEnum):
    """DS parallel types.

    - `REDUNDANT`:    Every process performs the computation redundantly.
    - `SYNCHRONIZED`: The first process sends the result to the rest.
    - `DISTRIBUTED`:  Used in some cases to distribute the computation among
                      processes.
    """

    REDUNDANT = ...
    """Every process performs the computation redundantly."""
    SYNCHRONIZED = ...
    """The first process sends the result to the rest."""
    DISTRIBUTED = ...
    """Used in some cases to distribute the computation among processes."""

class DS:
    """Direct solver for dense eigenproblems.

    DS is used internally by SLEPc solvers to handle the dense subproblems
    that appear in iterative methods.

    See Also
    --------
    slepc.DS
    """

    Type = DSType
    StateType = DSStateType
    MatType = DSMatType
    ParallelType = DSParallelType

    # --- View and Lifecycle ---

    def view(self, viewer: Viewer | None = None) -> None:
        """Print the DS data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.

        See Also
        --------
        slepc.DSView
        """
        ...

    def destroy(self) -> Self:
        """Destroy the DS object.

        Collective.

        See Also
        --------
        slepc.DSDestroy
        """
        ...

    def reset(self) -> None:
        """Reset the DS object.

        Collective.

        See Also
        --------
        slepc.DSReset
        """
        ...

    def create(self, comm: Comm | None = None) -> Self:
        """Create the DS object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator; if not provided, it defaults to all processes.

        See Also
        --------
        slepc.DSCreate
        """
        ...

    # --- Type and Options ---

    def setType(self, ds_type: DSType | str) -> None:
        """Set the type for the DS object.

        Logically collective.

        Parameters
        ----------
        ds_type
            The direct solver type to be used.

        See Also
        --------
        slepc.DSSetType
        """
        ...

    def getType(self) -> str:
        """Get the DS type of this object.

        Not collective.

        Returns
        -------
        str
            The direct solver type currently being used.

        See Also
        --------
        slepc.DSGetType
        """
        ...

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """Set the prefix used for searching for all DS options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all DS option requests.

        Notes
        -----
        A hyphen (``-``) must NOT be given at the beginning of the
        prefix name.  The first character of all runtime options is
        AUTOMATICALLY the hyphen.

        See Also
        --------
        slepc.DSSetOptionsPrefix
        """
        ...

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """Append to the prefix used for searching for all DS options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all DS option requests.

        See Also
        --------
        slepc.DSAppendOptionsPrefix
        """
        ...

    def getOptionsPrefix(self) -> str:
        """Get the prefix used for searching for all DS options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this DS object.

        See Also
        --------
        slepc.DSGetOptionsPrefix
        """
        ...

    def setFromOptions(self) -> None:
        """Set DS options from the options database.

        Collective.

        Notes
        -----
        To see all options, run your program with the ``-help``
        option.

        See Also
        --------
        slepc.DSSetFromOptions
        """
        ...

    def duplicate(self) -> DS:
        """Duplicate the DS object with the same type and dimensions.

        Collective.

        See Also
        --------
        slepc.DSDuplicate
        """
        ...

    # --- Allocation and Dimensions ---

    def allocate(self, ld: int) -> None:
        """Allocate memory for internal storage or matrices in DS.

        Logically collective.

        Parameters
        ----------
        ld
            Leading dimension (maximum allowed dimension for the
            matrices, including the extra row if present).

        See Also
        --------
        slepc.DSAllocate
        """
        ...

    def getLeadingDimension(self) -> int:
        """Get the leading dimension of the allocated matrices.

        Not collective.

        Returns
        -------
        int
            Leading dimension (maximum allowed dimension for the matrices).

        See Also
        --------
        slepc.DSGetLeadingDimension
        """
        ...

    def setState(self, state: DSStateType) -> None:
        """Set the state of the DS object.

        Logically collective.

        Parameters
        ----------
        state
            The new state.

        Notes
        -----
        The state indicates that the dense system is in an initial
        state (raw), in an intermediate state (such as tridiagonal,
        Hessenberg or Hessenberg-triangular), in a condensed state
        (such as diagonal, Schur or generalized Schur), or in a
        truncated state.

        This function is normally used to return to the raw state when
        the condensed structure is destroyed.

        See Also
        --------
        slepc.DSSetState
        """
        ...

    def getState(self) -> DSStateType:
        """Get the current state.

        Not collective.

        Returns
        -------
        DSStateType
            The current state.

        See Also
        --------
        slepc.DSGetState
        """
        ...

    def setParallel(self, pmode: DSParallelType) -> None:
        """Set the mode of operation in parallel runs.

        Logically collective.

        Parameters
        ----------
        pmode
            The parallel mode.

        See Also
        --------
        slepc.DSSetParallel
        """
        ...

    def getParallel(self) -> DSParallelType:
        """Get the mode of operation in parallel runs.

        Not collective.

        Returns
        -------
        DSParallelType
            The parallel mode.

        See Also
        --------
        slepc.DSGetParallel
        """
        ...

    def setDimensions(
        self,
        n: int | None = None,
        l: int | None = None,
        k: int | None = None,
    ) -> None:
        """Set the matrices sizes in the DS object.

        Logically collective.

        Parameters
        ----------
        n
            The new size.
        l
            Number of locked (inactive) leading columns.
        k
            Intermediate dimension (e.g., position of arrow).

        Notes
        -----
        The internal arrays are not reallocated.

        See Also
        --------
        slepc.DSSetDimensions
        """
        ...

    def getDimensions(self) -> tuple[int, int, int, int]:
        """Get the current dimensions.

        Not collective.

        Returns
        -------
        n : int
            The new size.
        l : int
            Number of locked (inactive) leading columns.
        k : int
            Intermediate dimension (e.g., position of arrow).
        t : int
            Truncated length.

        See Also
        --------
        slepc.DSGetDimensions
        """
        ...

    def setBlockSize(self, bs: int) -> None:
        """Set the block size.

        Logically collective.

        Parameters
        ----------
        bs
            The block size.

        See Also
        --------
        slepc.DSSetBlockSize
        """
        ...

    def getBlockSize(self) -> int:
        """Get the block size.

        Not collective.

        Returns
        -------
        int
            The block size.

        See Also
        --------
        slepc.DSGetBlockSize
        """
        ...

    def setMethod(self, meth: int) -> None:
        """Set the method to be used to solve the problem.

        Logically collective.

        Parameters
        ----------
        meth
            An index identifying the method.

        See Also
        --------
        slepc.DSSetMethod
        """
        ...

    def getMethod(self) -> int:
        """Get the method currently used in the DS.

        Not collective.

        Returns
        -------
        int
            Identifier of the method.

        See Also
        --------
        slepc.DSGetMethod
        """
        ...

    # --- Compact and Extra Row ---

    def setCompact(self, comp: bool) -> None:
        """Set the matrices' compact storage flag.

        Logically collective.

        Parameters
        ----------
        comp
            True means compact storage.

        Notes
        -----
        Compact storage is used in some `DS` types such as
        `DS.Type.HEP` when the matrix is tridiagonal. This flag
        can be used to indicate whether the user provides the
        matrix entries via the compact form (the tridiagonal
        `DS.MatType.T`) or the non-compact one (`DS.MatType.A`).

        The default is ``False``.

        See Also
        --------
        slepc.DSSetCompact
        """
        ...

    def getCompact(self) -> bool:
        """Get the compact storage flag.

        Not collective.

        Returns
        -------
        bool
            The flag.

        See Also
        --------
        slepc.DSGetCompact
        """
        ...

    def setExtraRow(self, ext: bool) -> None:
        """Set a flag to indicate that the matrix has one extra row.

        Logically collective.

        Parameters
        ----------
        ext
            True if the matrix has extra row.

        Notes
        -----
        In Krylov methods it is useful that the matrix representing the direct
        solver has one extra row, i.e., has dimension (n+1) x n. If
        this flag is activated, all transformations applied to the right of the
        matrix also affect this additional row. In that case, (n+1)
        must be less or equal than the leading dimension.

        The default is ``False``.

        See Also
        --------
        slepc.DSSetExtraRow
        """
        ...

    def getExtraRow(self) -> bool:
        """Get the extra row flag.

        Not collective.

        Returns
        -------
        bool
            The flag.

        See Also
        --------
        slepc.DSGetExtraRow
        """
        ...

    def setRefined(self, ref: bool) -> None:
        """Set a flag to indicate that refined vectors must be computed.

        Logically collective.

        Parameters
        ----------
        ref
            True if refined vectors must be used.

        Notes
        -----
        Normally the vectors returned in `DS.MatType.X` are eigenvectors of
        the projected matrix. With this flag activated, `vectors()` will return
        the right singular vector of the smallest singular value of matrix
        At - theta I, where At is the extended (n+1) x n matrix and
        theta is the Ritz value.
        This is used in the refined Ritz approximation.

        The default is ``False``.

        See Also
        --------
        slepc.DSSetRefined
        """
        ...

    def getRefined(self) -> bool:
        """Get the refined vectors flag.

        Not collective.

        Returns
        -------
        bool
            The flag.

        See Also
        --------
        slepc.DSGetRefined
        """
        ...

    def truncate(self, n: int, trim: bool = False) -> None:
        """Truncate the system represented in the DS object.

        Logically collective.

        Parameters
        ----------
        n
            The new size.
        trim
            A flag to indicate if the factorization must be trimmed.

        See Also
        --------
        slepc.DSTruncate
        """
        ...

    def updateExtraRow(self) -> None:
        """Ensure that the extra row gets up-to-date after a call to `DS.solve()`.

        Logically collective.

        Perform all necessary operations so that the extra row gets up-to-date
        after a call to `DS.solve()`.

        See Also
        --------
        slepc.DSUpdateExtraRow
        """
        ...

    # --- Matrix Operations ---

    def getMat(self, matname: DSMatType) -> Mat:
        """Get the requested matrix as a sequential dense Mat object.

        Not collective.

        Parameters
        ----------
        matname
            The requested matrix.

        Returns
        -------
        Mat
            The matrix.

        See Also
        --------
        slepc.DSGetMat
        """
        ...

    def restoreMat(self, matname: DSMatType, mat: Mat) -> None:
        """Restore the previously seized matrix.

        Not collective.

        Parameters
        ----------
        matname
            The selected matrix.
        mat
            The matrix previously obtained with `getMat()`.

        See Also
        --------
        slepc.DSRestoreMat
        """
        ...

    def setIdentity(self, matname: DSMatType) -> None:
        """Set the identity on the active part of a matrix.

        Logically collective.

        Parameters
        ----------
        matname
            The requested matrix.

        See Also
        --------
        slepc.DSSetIdentity
        """
        ...

    # --- Solve ---

    def cond(self) -> float:
        """Compute the inf-norm condition number of the first matrix.

        Logically collective.

        Returns
        -------
        float
            Condition number.

        See Also
        --------
        slepc.DSCond
        """
        ...

    def solve(self) -> ArrayScalar:
        """Solve the problem.

        Logically collective.

        Returns
        -------
        ArrayScalar
            Eigenvalues or singular values.

        See Also
        --------
        slepc.DSSolve
        """
        ...

    def vectors(self, matname: DSMatType = ...) -> None:
        """Compute vectors associated to the dense system such as eigenvectors.

        Logically collective.

        Parameters
        ----------
        matname
            The matrix, used to indicate which vectors are required.
            Default is `DS.MatType.X`.

        See Also
        --------
        slepc.DSVectors
        """
        ...

    # --- SVD Specific ---

    def setSVDDimensions(self, m: int) -> None:
        """Set the number of columns of a `DS` of type `SVD`.

        Logically collective.

        Parameters
        ----------
        m
            The number of columns.

        See Also
        --------
        slepc.DSSVDSetDimensions
        """
        ...

    def getSVDDimensions(self) -> int:
        """Get the number of columns of a `DS` of type `SVD`.

        Not collective.

        Returns
        -------
        int
            The number of columns.

        See Also
        --------
        slepc.DSSVDGetDimensions
        """
        ...

    def setHSVDDimensions(self, m: int) -> None:
        """Set the number of columns of a `DS` of type `HSVD`.

        Logically collective.

        Parameters
        ----------
        m
            The number of columns.

        See Also
        --------
        slepc.DSHSVDSetDimensions
        """
        ...

    def getHSVDDimensions(self) -> int:
        """Get the number of columns of a `DS` of type `HSVD`.

        Not collective.

        Returns
        -------
        int
            The number of columns.

        See Also
        --------
        slepc.DSHSVDGetDimensions
        """
        ...

    def setGSVDDimensions(self, m: int, p: int) -> None:
        """Set the number of columns and rows of a `DS` of type `GSVD`.

        Logically collective.

        Parameters
        ----------
        m
            The number of columns.
        p
            The number of rows for the second matrix.

        See Also
        --------
        slepc.DSGSVDSetDimensions
        """
        ...

    def getGSVDDimensions(self) -> tuple[int, int]:
        """Get the number of columns and rows of a `DS` of type `GSVD`.

        Not collective.

        Returns
        -------
        m : int
            The number of columns.
        p : int
            The number of rows for the second matrix.

        See Also
        --------
        slepc.DSGSVDGetDimensions
        """
        ...

    # --- PEP Specific ---

    def setPEPDegree(self, deg: int) -> None:
        """Set the polynomial degree of a `DS` of type `PEP`.

        Logically collective.

        Parameters
        ----------
        deg
            The polynomial degree.

        See Also
        --------
        slepc.DSPEPSetDegree
        """
        ...

    def getPEPDegree(self) -> int:
        """Get the polynomial degree of a `DS` of type `PEP`.

        Not collective.

        Returns
        -------
        int
            The polynomial degree.

        See Also
        --------
        slepc.DSPEPGetDegree
        """
        ...

    def setPEPCoefficients(self, pbc: Sequence[float]) -> None:
        """Set the polynomial basis coefficients of a `DS` of type `PEP`.

        Logically collective.

        Parameters
        ----------
        pbc
            Coefficients.

        See Also
        --------
        slepc.DSPEPSetCoefficients
        """
        ...

    def getPEPCoefficients(self) -> ArrayReal:
        """Get the polynomial basis coefficients of a `DS` of type `PEP`.

        Not collective.

        Returns
        -------
        ArrayReal
            Coefficients.

        See Also
        --------
        slepc.DSPEPGetCoefficients
        """
        ...

    # --- Properties ---

    @property
    def state(self) -> DSStateType:
        """The state of the DS object."""
        ...

    @state.setter
    def state(self, value: DSStateType) -> None: ...
    @property
    def parallel(self) -> DSParallelType:
        """The mode of operation in parallel runs."""
        ...

    @parallel.setter
    def parallel(self, value: DSParallelType) -> None: ...
    @property
    def block_size(self) -> int:
        """The block size."""
        ...

    @block_size.setter
    def block_size(self, value: int) -> None: ...
    @property
    def method(self) -> int:
        """The method to be used to solve the problem."""
        ...

    @method.setter
    def method(self, value: int) -> None: ...
    @property
    def compact(self) -> bool:
        """Compact storage of matrices."""
        ...

    @compact.setter
    def compact(self, value: bool) -> None: ...
    @property
    def extra_row(self) -> bool:
        """If the matrix has one extra row."""
        ...

    @extra_row.setter
    def extra_row(self, value: bool) -> None: ...
    @property
    def refined(self) -> bool:
        """If refined vectors must be computed."""
        ...

    @refined.setter
    def refined(self, value: bool) -> None: ...



# -----------------------------------------------------------------------------

class MFNType(object):
    """
    MFN type.

    Action of a matrix function on a vector.

    - `KRYLOV`:  Restarted Krylov solver.
    - `EXPOKIT`: Implementation of the method in Expokit.
    """
    KRYLOV   = S_(MFNKRYLOV)
    EXPOKIT  = S_(MFNEXPOKIT)

class MFNConvergedReason(object):
    """
    MFN convergence reasons.

    - 'MFN_CONVERGED_TOL': All eigenpairs converged to requested tolerance.
    - 'MFN_CONVERGED_ITS': Solver completed the requested number of steps.
    - 'MFN_DIVERGED_ITS': Maximum number of iterations exceeded.
    - 'MFN_DIVERGED_BREAKDOWN': Generic breakdown in method.
    """
    CONVERGED_TOL       = MFN_CONVERGED_TOL
    CONVERGED_ITS       = MFN_CONVERGED_ITS
    DIVERGED_ITS        = MFN_DIVERGED_ITS
    DIVERGED_BREAKDOWN  = MFN_DIVERGED_BREAKDOWN
    CONVERGED_ITERATING = MFN_CONVERGED_ITERATING
    ITERATING           = MFN_CONVERGED_ITERATING

# -----------------------------------------------------------------------------

cdef class MFN(Object):

    """MFN."""

    Type            = MFNType
    ConvergedReason = MFNConvergedReason

    def __cinit__(self):
        self.obj = <PetscObject*> &self.mfn
        self.mfn = NULL

    def view(self, Viewer viewer=None) -> None:
        """
        Print the MFN data structure.

        Collective.

        Parameters
        ----------
        viewer
            Visualization context; if not provided, the standard
            output is used.
        """
        cdef PetscViewer vwr = def_Viewer(viewer)
        CHKERR( MFNView(self.mfn, vwr) )

    def destroy(self) -> Self:
        """
        Destroy the MFN object.

        Logically collective.
        """
        CHKERR( MFNDestroy(&self.mfn) )
        self.mfn = NULL
        return self

    def reset(self) -> None:
        """
        Reset the MFN object.

        Collective.
        """
        CHKERR( MFNReset(self.mfn) )

    def create(self, comm: Comm | None = None) -> Self:
        """
        Create the MFN object.

        Collective.

        Parameters
        ----------
        comm
            MPI communicator. If not provided, it defaults to all processes.
        """
        cdef MPI_Comm ccomm = def_Comm(comm, SLEPC_COMM_DEFAULT())
        cdef SlepcMFN newmfn = NULL
        CHKERR( MFNCreate(ccomm, &newmfn) )
        CHKERR( SlepcCLEAR(self.obj) ); self.mfn = newmfn
        return self

    def setType(self, mfn_type: Type | str) -> None:
        """
        Set the particular solver to be used in the MFN object.

        Logically collective.

        Parameters
        ----------
        mfn_type
            The solver to be used.
        """
        cdef SlepcMFNType cval = NULL
        mfn_type = str2bytes(mfn_type, &cval)
        CHKERR( MFNSetType(self.mfn, cval) )

    def getType(self) -> str:
        """
        Get the MFN type of this object.

        Not collective.

        Returns
        -------
        str
            The solver currently being used.
        """
        cdef SlepcMFNType mfn_type = NULL
        CHKERR( MFNGetType(self.mfn, &mfn_type) )
        return bytes2str(mfn_type)

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for searching for all MFN options in the database.

        Not collective.

        Returns
        -------
        str
            The prefix string set for this MFN object.
        """
        cdef const char *prefix = NULL
        CHKERR( MFNGetOptionsPrefix(self.mfn, &prefix) )
        return bytes2str(prefix)

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for searching for all MFN options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all MFN option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( MFNSetOptionsPrefix(self.mfn, cval) )

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for searching for all MFN options in the database.

        Logically collective.

        Parameters
        ----------
        prefix
            The prefix string to prepend to all MFN option requests.
        """
        cdef const char *cval = NULL
        prefix = str2bytes(prefix, &cval)
        CHKERR( MFNAppendOptionsPrefix(self.mfn, cval) )

    def setFromOptions(self) -> None:
        """
        Set MFN options from the options database.

        Collective.

        Set MFN options from the options database. This routine must
        be called before `setUp()` if the user is to be allowed to set
        the solver type.
        """
        CHKERR( MFNSetFromOptions(self.mfn) )

    def getTolerances(self) -> tuple[float, int]:
        """
        Get the tolerance and maximum iteration count.

        Not collective.

        Get the tolerance and maximum iteration count used by the
        default MFN convergence tests.

        Returns
        -------
        tol: float
            The convergence tolerance.
        max_it: int
            The maximum number of iterations
        """
        cdef PetscReal rval = 0
        cdef PetscInt  ival = 0
        CHKERR( MFNGetTolerances(self.mfn, &rval, &ival) )
        return (toReal(rval), toInt(ival))

    def setTolerances(self, tol: float | None = None, max_it: int | None = None) -> None:
        """
        Set the tolerance and maximum iteration count.

        Logically collective.

        Set the tolerance and maximum iteration count used by the
        default MFN convergence tests.

        Parameters
        ----------
        tol
            The convergence tolerance.
        max_it
            The maximum number of iterations
        """
        cdef PetscReal rval = PETSC_CURRENT
        cdef PetscInt  ival = PETSC_CURRENT
        if tol    is not None: rval = asReal(tol)
        if max_it is not None: ival = asInt(max_it)
        CHKERR( MFNSetTolerances(self.mfn, rval, ival) )

    def getDimensions(self) -> int:
        """
        Get the dimension of the subspace used by the solver.

        Not collective.

        Returns
        -------
        int
            Maximum dimension of the subspace to be used by the solver.
        """
        cdef PetscInt ival = 0
        CHKERR( MFNGetDimensions(self.mfn, &ival) )
        return toInt(ival)

    def setDimensions(self, ncv: int) -> None:
        """
        Set the dimension of the subspace to be used by the solver.

        Logically collective.

        Parameters
        ----------
        ncv
            Maximum dimension of the subspace to be used by the solver.
        """
        cdef PetscInt ival = asInt(ncv)
        CHKERR( MFNSetDimensions(self.mfn, ival) )

    def getFN(self) -> FN:
        """
        Get the math function object associated to the MFN object.

        Not collective.

        Returns
        -------
        FN
            The math function context.
        """
        cdef FN fn = FN()
        CHKERR( MFNGetFN(self.mfn, &fn.fn) )
        CHKERR( PetscINCREF(fn.obj) )
        return fn

    def setFN(self, FN fn) -> None:
        """
        Set a math function object associated to the MFN object.

        Collective.

        Parameters
        ----------
        fn
            The math function context.
        """
        CHKERR( MFNSetFN(self.mfn, fn.fn) )

    def getBV(self) -> BV:
        """
        Get the basis vector object associated to the MFN object.

        Not collective.

        Returns
        -------
        BV
            The basis vectors context.
        """
        cdef BV bv = BV()
        CHKERR( MFNGetBV(self.mfn, &bv.bv) )
        CHKERR( PetscINCREF(bv.obj) )
        return bv

    def setBV(self, BV bv) -> None:
        """
        Set a basis vector object associated to the MFN object.

        Collective.

        Parameters
        ----------
        bv
            The basis vectors context.
        """
        CHKERR( MFNSetBV(self.mfn, bv.bv) )

    def getOperator(self) -> petsc4py.PETSc.Mat:
        """
        Get the matrix associated with the MFN object.

        Collective.

        Returns
        -------
        petsc4py.PETSc.Mat
            The matrix for which the matrix function is to be computed.
        """
        cdef Mat A = Mat()
        CHKERR( MFNGetOperator(self.mfn, &A.mat) )
        CHKERR( PetscINCREF(A.obj) )
        return A

    def setOperator(self, Mat A) -> None:
        """
        Set the matrix associated with the MFN object.

        Collective.

        Parameters
        ----------
        A
            The problem matrix.
        """
        CHKERR( MFNSetOperator(self.mfn, A.mat) )

    #

    def setMonitor(
        self,
        monitor: MFNMonitorFunction | None,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Append a monitor function to the list of monitors.

        Logically collective.
        """
        if monitor is None: return
        cdef object monitorlist = self.get_attr('__monitor__')
        if monitorlist is None:
            monitorlist = []
            self.set_attr('__monitor__', monitorlist)
            CHKERR( MFNMonitorSet(self.mfn, MFN_Monitor, NULL, NULL) )
        if args is None: args = ()
        if kargs is None: kargs = {}
        monitorlist.append((monitor, args, kargs))

    def getMonitor(self) -> MFNMonitorFunction:
        """Get the list of monitor functions."""
        return self.get_attr('__monitor__')

    def cancelMonitor(self) -> None:
        """
        Clear all monitors for an `MFN` object.

        Logically collective.
        """
        CHKERR( MFNMonitorCancel(self.mfn) )
        self.set_attr('__monitor__', None)

    #

    def setUp(self) -> None:
        """
        Set up all the necessary internal data structures.

        Collective.

        Set up all the internal data structures necessary for the execution
        of the eigensolver.
        """
        CHKERR( MFNSetUp(self.mfn) )

    def solve(self, Vec b, Vec x) -> None:
        """
        Solve the matrix function problem.

        Collective.

        Given a vector :math:`b`, the vector :math:`x = f(A) b` is
        returned.

        Parameters
        ----------
        b
            The right hand side vector.
        x
            The solution.
        """
        CHKERR( MFNSolve(self.mfn, b.vec, x.vec) )

    def solveTranspose(self, Vec b, Vec x) -> None:
        """
        Solve the transpose matrix function problem.

        Collective.

        Given a vector :math:`b`, the vector :math:`x = f(A^T) b` is
        returned.

        Parameters
        ----------
        b
            The right hand side vector.
        x
            The solution.
        """
        CHKERR( MFNSolveTranspose(self.mfn, b.vec, x.vec) )

    def getIterationNumber(self) -> int:
        """
        Get the current iteration number.

        Not collective.

        Get the current iteration number. If the call to `solve()` is
        complete, then it returns the number of iterations carried out
        by the solution method.

        Returns
        -------
        int
            Iteration number.
        """
        cdef PetscInt ival = 0
        CHKERR( MFNGetIterationNumber(self.mfn, &ival) )
        return toInt(ival)

    def getConvergedReason(self) -> ConvergedReason:
        """
        Get the reason why the `solve()` iteration was stopped.

        Not collective.

        Returns
        -------
        ConvergedReason
            Negative value indicates diverged, positive value converged.
        """
        cdef SlepcMFNConvergedReason val = MFN_CONVERGED_ITERATING
        CHKERR( MFNGetConvergedReason(self.mfn, &val) )
        return val

    def setErrorIfNotConverged(self, flg: bool = True) -> None:
        """
        Set `solve()` to generate an error if the solver does not converge.

        Logically collective.

        Parameters
        ----------
        flg
            True indicates you want the error generated.
        """
        cdef PetscBool tval = flg
        CHKERR( MFNSetErrorIfNotConverged(self.mfn, tval) )

    def getErrorIfNotConverged(self) -> bool:
        """
        Get if `solve()` generates an error if the solver does not converge.

        Not collective.

        Get a flag indicating whether `solve()` will generate an error if the
        solver does not converge.

        Returns
        -------
        bool
            True indicates you want the error generated.
        """
        cdef PetscBool tval = PETSC_FALSE
        CHKERR( MFNGetErrorIfNotConverged(self.mfn, &tval) )
        return toBool(tval)

    #

    property tol:
        """The tolerance count used by the MFN convergence tests."""
        def __get__(self) -> float:
            return self.getTolerances()[0]
        def __set__(self, value):
            self.setTolerances(tol=value)

    property max_it:
        """The maximum iteration count used by the MFN convergence tests."""
        def __get__(self) -> int:
            return self.getTolerances()[1]
        def __set__(self, value):
            self.setTolerances(max_it=value)

    property fn:
        """The math function (FN) object associated to the MFN object."""
        def __get__(self) -> FN:
            return self.getFN()
        def __set__(self, value):
            self.setBV(value)

    property bv:
        """The basis vectors (BV) object associated to the MFN object."""
        def __get__(self) -> BV:
            return self.getFN()
        def __set__(self, value):
            self.setBV(value)

# -----------------------------------------------------------------------------

del MFNType
del MFNConvergedReason

# -----------------------------------------------------------------------------



cdef extern from * nogil:

    ctypedef char* SlepcRGType "const char*"
    SlepcRGType RGINTERVAL
    SlepcRGType RGPOLYGON
    SlepcRGType RGELLIPSE
    SlepcRGType RGRING

    ctypedef enum SlepcRGQuadRule "RGQuadRule":
        RG_QUADRULE_TRAPEZOIDAL
        RG_QUADRULE_CHEBYSHEV

    PetscErrorCode RGCreate(MPI_Comm,SlepcRG*)
    PetscErrorCode RGView(SlepcRG,PetscViewer)
    PetscErrorCode RGDestroy(SlepcRG*)
    PetscErrorCode RGSetType(SlepcRG,SlepcRGType)
    PetscErrorCode RGGetType(SlepcRG,SlepcRGType*)

    PetscErrorCode RGSetOptionsPrefix(SlepcRG,char[])
    PetscErrorCode RGGetOptionsPrefix(SlepcRG,char*[])
    PetscErrorCode RGAppendOptionsPrefix(SlepcRG,char[])
    PetscErrorCode RGSetFromOptions(SlepcRG)

    PetscErrorCode RGIsTrivial(SlepcRG,PetscBool*)
    PetscErrorCode RGIsAxisymmetric(SlepcRG,PetscBool,PetscBool*)
    PetscErrorCode RGSetComplement(SlepcRG,PetscBool)
    PetscErrorCode RGGetComplement(SlepcRG,PetscBool*)

    PetscErrorCode RGSetScale(SlepcRG,PetscReal)
    PetscErrorCode RGGetScale(SlepcRG,PetscReal*)
    PetscErrorCode RGPushScale(SlepcRG,PetscReal)
    PetscErrorCode RGPopScale(SlepcRG)

    PetscErrorCode RGCheckInside(SlepcRG,PetscInt,PetscScalar*,PetscScalar*,PetscInt*)
    PetscErrorCode RGComputeContour(SlepcRG,PetscInt,PetscScalar*,PetscScalar*)
    PetscErrorCode RGComputeBoundingBox(SlepcRG,PetscReal*,PetscReal*,PetscReal*,PetscReal*)
    PetscErrorCode RGCanUseConjugates(SlepcRG,PetscBool,PetscBool*)
    PetscErrorCode RGComputeQuadrature(SlepcRG,SlepcRGQuadRule,PetscInt,PetscScalar*,PetscScalar*,PetscScalar*)

    PetscErrorCode RGEllipseSetParameters(SlepcRG,PetscScalar,PetscReal,PetscReal)
    PetscErrorCode RGEllipseGetParameters(SlepcRG,PetscScalar*,PetscReal*,PetscReal*)
    PetscErrorCode RGIntervalSetEndpoints(SlepcRG,PetscReal,PetscReal,PetscReal,PetscReal)
    PetscErrorCode RGIntervalGetEndpoints(SlepcRG,PetscReal*,PetscReal*,PetscReal*,PetscReal*)
    PetscErrorCode RGPolygonSetVertices(SlepcRG,PetscInt,PetscScalar*,PetscScalar*)
    PetscErrorCode RGPolygonGetVertices(SlepcRG,PetscInt*,PetscScalar**,PetscScalar**)
    PetscErrorCode RGRingSetParameters(SlepcRG,PetscScalar,PetscReal,PetscReal,PetscReal,PetscReal,PetscReal)
    PetscErrorCode RGRingGetParameters(SlepcRG,PetscScalar*,PetscReal*,PetscReal*,PetscReal*,PetscReal*,PetscReal*)




"""Type stubs for SLEPc SVD module."""

from enum import IntEnum, StrEnum
from typing import Any, Callable, Sequence

from petsc4py.PETSc import KSP, Comm, Mat, Vec, Viewer

from .BV import BV
from .DS import DS
from .EPS import EPS

class SVDType(StrEnum):
    """SVD types."""

    CROSS = ...
    """Cross product."""
    CYCLIC = ...
    """Cyclic matrix."""
    LAPACK = ...
    """LAPACK solver."""
    LANCZOS = ...
    """Lanczos."""
    TRLANCZOS = ...
    """Thick-restart Lanczos."""
    RANDOMIZED = ...
    """Randomized SVD."""
    SCALAPACK = ...
    """ScaLAPACK solver."""
    KSVD = ...
    """KSVD solver."""
    ELEMENTAL = ...
    """Elemental solver."""
    PRIMME = ...
    """PRIMME solver."""

class SVDProblemType(IntEnum):
    """SVD problem type."""

    STANDARD = ...
    """Standard SVD."""
    GENERALIZED = ...
    """Generalized SVD."""
    HYPERBOLIC = ...
    """Hyperbolic SVD."""

class SVDErrorType(IntEnum):
    """SVD error type to assess accuracy of computed solutions."""

    ABSOLUTE = ...
    """Absolute error."""
    RELATIVE = ...
    """Relative error."""
    NORM = ...
    """Norm-based error."""

class SVDWhich(IntEnum):
    """SVD desired part of spectrum."""

    LARGEST = ...
    """Largest singular values."""
    SMALLEST = ...
    """Smallest singular values."""

class SVDConv(IntEnum):
    """SVD convergence test."""

    ABS = ...
    """Absolute convergence test."""
    REL = ...
    """Relative convergence test."""
    NORM = ...
    """Norm-based convergence test."""
    MAXIT = ...
    """Maximum iterations."""
    USER = ...
    """User-defined convergence test."""

class SVDStop(IntEnum):
    """SVD stopping test."""

    BASIC = ...
    """Default stopping test."""
    USER = ...
    """User-defined stopping test."""
    THRESHOLD = ...
    """Threshold stopping test."""

class SVDConvergedReason(IntEnum):
    """SVD convergence reasons."""

    CONVERGED_TOL = ...
    """Converged to requested tolerance."""
    CONVERGED_USER = ...
    """User-defined convergence criterion satisfied."""
    CONVERGED_MAXIT = ...
    """Converged within maximum iterations."""
    DIVERGED_ITS = ...
    """Maximum number of iterations exceeded."""
    DIVERGED_BREAKDOWN = ...
    """Solver failed due to breakdown."""
    DIVERGED_SYMMETRY_LOST = ...
    """Solver could not preserve symmetry."""
    CONVERGED_ITERATING = ...
    """Iteration not finished yet."""
    ITERATING = ...
    """Iteration not finished yet (alias)."""

class SVDTRLanczosGBidiag(IntEnum):
    """SVD TRLanczos bidiagonalization choices for the GSVD case."""

    SINGLE = ...
    """Single bidiagonalization."""
    UPPER = ...
    """Upper bidiagonalization."""
    LOWER = ...
    """Lower bidiagonalization."""

class SVD:
    """Singular Value Decomposition solver."""

    Type: type[SVDType]
    ProblemType: type[SVDProblemType]
    ErrorType: type[SVDErrorType]
    Which: type[SVDWhich]
    Conv: type[SVDConv]
    Stop: type[SVDStop]
    ConvergedReason: type[SVDConvergedReason]
    TRLanczosGBidiag: type[SVDTRLanczosGBidiag]

    def __init__(self) -> None: ...

    # --- Basic operations ---

    def view(self, viewer: Viewer | None = None) -> None:
        """
        Print the SVD data structure.

        Parameters
        ----------
        viewer
            Viewer to print the SVD object.
        """
        ...

    def destroy(self) -> SVD:
        """Destroy the SVD object."""
        ...

    def reset(self) -> None:
        """Reset the SVD object to its initial state."""
        ...

    def create(self, comm: Comm | None = None) -> SVD:
        """
        Create the SVD object.

        Parameters
        ----------
        comm
            MPI communicator.

        Returns
        -------
        SVD
            The created SVD object.
        """
        ...

    def setType(self, svd_type: SVDType | str) -> None:
        """
        Set the type of the SVD object.

        Parameters
        ----------
        svd_type
            The type of SVD solver to use.
        """
        ...

    def getType(self) -> str:
        """
        Get the type of the SVD object.

        Returns
        -------
        str
            The type of SVD solver.
        """
        ...

    def getOptionsPrefix(self) -> str:
        """
        Get the prefix used for all SVD options in the database.

        Returns
        -------
        str
            The options prefix string.
        """
        ...

    def setOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Set the prefix used for all SVD options in the database.

        Parameters
        ----------
        prefix
            The prefix string.
        """
        ...

    def appendOptionsPrefix(self, prefix: str | None = None) -> None:
        """
        Append to the prefix used for all SVD options in the database.

        Parameters
        ----------
        prefix
            The prefix string to append.
        """
        ...

    def setFromOptions(self) -> None:
        """Set SVD options from the options database."""
        ...

    # --- Problem type ---

    def getProblemType(self) -> SVDProblemType:
        """
        Get the problem type from the SVD object.

        Returns
        -------
        SVDProblemType
            The problem type.
        """
        ...

    def setProblemType(self, problem_type: SVDProblemType) -> None:
        """
        Set the type of the singular value problem.

        Parameters
        ----------
        problem_type
            The problem type to be set.
        """
        ...

    def isGeneralized(self) -> bool:
        """
        Tell if the SVD corresponds to a generalized singular value problem.

        Returns
        -------
        bool
            True if two matrices were set with setOperators().
        """
        ...

    def isHyperbolic(self) -> bool:
        """
        Tell whether the SVD object corresponds to a hyperbolic singular value problem.

        Returns
        -------
        bool
            True if the problem was specified as hyperbolic.
        """
        ...

    # --- Transpose mode ---

    def getImplicitTranspose(self) -> bool:
        """
        Get the mode used to handle the transpose of the matrix.

        Returns
        -------
        bool
            How to handle the transpose (implicitly or not).
        """
        ...

    def setImplicitTranspose(self, mode: bool) -> None:
        """
        Set how to handle the transpose of the matrix.

        Parameters
        ----------
        mode
            How to handle the transpose (implicitly or not).
        """
        ...

    # --- Spectrum selection ---

    def getWhichSingularTriplets(self) -> SVDWhich:
        """
        Get which singular triplets are to be sought.

        Returns
        -------
        SVDWhich
            The singular values to be sought (either largest or smallest).
        """
        ...

    def setWhichSingularTriplets(self, which: SVDWhich) -> None:
        """
        Set which singular triplets are to be sought.

        Parameters
        ----------
        which
            The singular values to be sought (either largest or smallest).
        """
        ...

    # --- Threshold ---

    def getThreshold(self) -> tuple[float, bool]:
        """
        Get the threshold used in the threshold stopping test.

        Returns
        -------
        thres : float
            The threshold.
        rel : bool
            Whether the threshold is relative or not.
        """
        ...

    def setThreshold(self, thres: float, rel: bool = False) -> None:
        """
        Set the threshold used in the threshold stopping test.

        Parameters
        ----------
        thres
            The threshold.
        rel
            Whether the threshold is relative or not.
        """
        ...

    # --- Tolerances and convergence ---

    def getTolerances(self) -> tuple[float, int]:
        """
        Get the tolerance and maximum iteration count.

        Returns
        -------
        tol : float
            The convergence tolerance.
        max_it : int
            Maximum number of iterations.
        """
        ...

    def setTolerances(
        self, tol: float | None = None, max_it: int | None = None
    ) -> None:
        """
        Set the tolerance and maximum iteration count.

        Parameters
        ----------
        tol
            The convergence tolerance.
        max_it
            Maximum number of iterations.
        """
        ...

    def getConvergenceTest(self) -> SVDConv:
        """
        Get the method used to compute the error estimate used in convergence test.

        Returns
        -------
        SVDConv
            The convergence test method.
        """
        ...

    def setConvergenceTest(self, conv: SVDConv) -> None:
        """
        Set the convergence test method.

        Parameters
        ----------
        conv
            The convergence test method.
        """
        ...

    def getTrackAll(self) -> bool:
        """
        Get the flag indicating whether all residual norms must be computed.

        Returns
        -------
        bool
            Whether all residuals are computed.
        """
        ...

    def setTrackAll(self, trackall: bool) -> None:
        """
        Set flag to compute the residual of all singular triplets.

        Parameters
        ----------
        trackall
            Whether to compute all residuals.
        """
        ...

    # --- Dimensions ---

    def getDimensions(self) -> tuple[int, int, int]:
        """
        Get the number of singular values to compute and the dimension of the subspace.

        Returns
        -------
        nsv : int
            Number of singular values to compute.
        ncv : int
            Maximum dimension of the subspace.
        mpd : int
            Maximum dimension allowed for the projected problem.
        """
        ...

    def setDimensions(
        self,
        nsv: int | None = None,
        ncv: int | None = None,
        mpd: int | None = None,
    ) -> None:
        """
        Set the number of singular values to compute and the dimension of the subspace.

        Parameters
        ----------
        nsv
            Number of singular values to compute.
        ncv
            Maximum dimension of the subspace.
        mpd
            Maximum dimension allowed for the projected problem.
        """
        ...

    # --- Associated objects ---

    def getBV(self) -> tuple[BV, BV]:
        """
        Get the basis vectors objects associated with the SVD object.

        Returns
        -------
        V : BV
            The basis vectors context for right singular vectors.
        U : BV
            The basis vectors context for left singular vectors.
        """
        ...

    def setBV(self, V: BV, U: BV | None = None) -> None:
        """
        Set basis vectors objects associated with the SVD solver.

        Parameters
        ----------
        V
            The basis vectors context for right singular vectors.
        U
            The basis vectors context for left singular vectors.
        """
        ...

    def getDS(self) -> DS:
        """
        Get the direct solver associated with the singular value solver.

        Returns
        -------
        DS
            The direct solver context.
        """
        ...

    def setDS(self, ds: DS) -> None:
        """
        Set a direct solver object associated with the singular value solver.

        Parameters
        ----------
        ds
            The direct solver context.
        """
        ...

    # --- Operators ---

    def getOperators(self) -> tuple[Mat, Mat | None]:
        """
        Get the matrices associated with the singular value problem.

        Returns
        -------
        A : Mat
            The matrix associated with the singular value problem.
        B : Mat or None
            The second matrix in the case of GSVD.
        """
        ...

    def setOperators(self, A: Mat, B: Mat | None = None) -> None:
        """
        Set the matrices associated with the singular value problem.

        Parameters
        ----------
        A
            The matrix associated with the singular value problem.
        B
            The second matrix in the case of GSVD.
        """
        ...

    def setOperator(self, A: Mat, B: Mat | None = None) -> None:
        """
        Set the matrices associated with the singular value problem.

        Parameters
        ----------
        A
            The matrix associated with the singular value problem.
        B
            The second matrix in the case of GSVD.
        """
        ...

    def getSignature(self, omega: Vec | None = None) -> Vec:
        """
        Get the signature matrix defining a hyperbolic singular value problem.

        Parameters
        ----------
        omega
            Optional vector to store the diagonal elements.

        Returns
        -------
        Vec
            A vector containing the diagonal elements of the signature matrix.
        """
        ...

    def setSignature(self, omega: Vec | None = None) -> None:
        """
        Set the signature matrix defining a hyperbolic singular value problem.

        Parameters
        ----------
        omega
            A vector containing the diagonal elements of the signature matrix.
        """
        ...

    # --- Initial space and callbacks ---

    def setInitialSpace(
        self,
        spaceright: Sequence[Vec] | None = None,
        spaceleft: Sequence[Vec] | None = None,
    ) -> None:
        """
        Set the initial spaces from which the SVD solver starts to iterate.

        Parameters
        ----------
        spaceright
            The right initial space.
        spaceleft
            The left initial space.
        """
        ...

    def setStoppingTest(
        self,
        stopping: Callable[..., SVDConvergedReason] | None,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Set a function to decide when to stop the outer iteration.

        Parameters
        ----------
        stopping
            A callable to test for stopping.
        args
            Positional arguments for the callable.
        kargs
            Keyword arguments for the callable.
        """
        ...

    def getStoppingTest(self) -> Callable[..., SVDConvergedReason] | None:
        """
        Get the stopping function.

        Returns
        -------
        Callable or None
            The stopping test function.
        """
        ...

    def setMonitor(
        self,
        monitor: Callable[..., None] | None,
        args: tuple[Any, ...] | None = None,
        kargs: dict[str, Any] | None = None,
    ) -> None:
        """
        Append a monitor function to the list of monitors.

        Parameters
        ----------
        monitor
            A callable to monitor the solver.
        args
            Positional arguments for the callable.
        kargs
            Keyword arguments for the callable.
        """
        ...

    def getMonitor(self) -> list[tuple[Callable[..., None], tuple, dict]] | None:
        """
        Get the list of monitor functions.

        Returns
        -------
        list or None
            The list of monitor functions.
        """
        ...

    def cancelMonitor(self) -> None:
        """Clear all monitors for the SVD object."""
        ...

    # --- Solve ---

    def setUp(self) -> None:
        """Set up all the necessary internal data structures."""
        ...

    def solve(self) -> None:
        """Solve the singular value problem."""
        ...

    def getIterationNumber(self) -> int:
        """
        Get the current iteration number.

        Returns
        -------
        int
            Iteration number.
        """
        ...

    def getConvergedReason(self) -> SVDConvergedReason:
        """
        Get the reason why the solve iteration was stopped.

        Returns
        -------
        SVDConvergedReason
            Negative value indicates diverged, positive value converged.
        """
        ...

    def getConverged(self) -> int:
        """
        Get the number of converged singular triplets.

        Returns
        -------
        int
            Number of converged singular triplets.
        """
        ...

    def getValue(self, i: int) -> float:
        """
        Get the i-th singular value as computed by solve().

        Parameters
        ----------
        i
            Index of the solution to be obtained.

        Returns
        -------
        float
            The computed singular value.
        """
        ...

    def getVectors(self, i: int, U: Vec, V: Vec) -> None:
        """
        Get the i-th left and right singular vectors as computed by solve().

        Parameters
        ----------
        i
            Index of the solution to be obtained.
        U
            Placeholder for the returned left singular vector.
        V
            Placeholder for the returned right singular vector.
        """
        ...

    def getSingularTriplet(
        self, i: int, U: Vec | None = None, V: Vec | None = None
    ) -> float:
        """
        Get the i-th triplet of the singular value decomposition.

        Parameters
        ----------
        i
            Index of the solution to be obtained.
        U
            Placeholder for the returned left singular vector.
        V
            Placeholder for the returned right singular vector.

        Returns
        -------
        float
            The computed singular value.
        """
        ...

    def computeError(self, i: int, etype: SVDErrorType | None = None) -> float:
        """
        Compute the error associated with the i-th singular triplet.

        Parameters
        ----------
        i
            Index of the solution.
        etype
            The error type.

        Returns
        -------
        float
            The error bound.
        """
        ...

    # --- View methods ---

    def errorView(
        self, etype: SVDErrorType | None = None, viewer: Viewer | None = None
    ) -> None:
        """
        Display the errors associated with the computed solution.

        Parameters
        ----------
        etype
            The error type.
        viewer
            Visualization context.
        """
        ...

    def valuesView(self, viewer: Viewer | None = None) -> None:
        """
        Display the computed singular values.

        Parameters
        ----------
        viewer
            Visualization context.
        """
        ...

    def vectorsView(self, viewer: Viewer | None = None) -> None:
        """
        Output computed singular vectors to a viewer.

        Parameters
        ----------
        viewer
            Visualization context.
        """
        ...

    # === Cross-specific methods ===

    def setCrossEPS(self, eps: EPS) -> None:
        """
        Set an eigensolver object associated with the singular value solver.

        Parameters
        ----------
        eps
            The eigensolver object.
        """
        ...

    def getCrossEPS(self) -> EPS:
        """
        Get the eigensolver object associated with the singular value solver.

        Returns
        -------
        EPS
            The eigensolver object.
        """
        ...

    def setCrossExplicitMatrix(self, flag: bool = True) -> None:
        """
        Set if the eigensolver operator A^T A must be computed explicitly.

        Parameters
        ----------
        flag
            True to build A^T A explicitly.
        """
        ...

    def getCrossExplicitMatrix(self) -> bool:
        """
        Get the flag indicating if A^T*A is built explicitly.

        Returns
        -------
        bool
            True if A^T*A is built explicitly.
        """
        ...

    # === Cyclic-specific methods ===

    def setCyclicEPS(self, eps: EPS) -> None:
        """
        Set an eigensolver object associated with the singular value solver.

        Parameters
        ----------
        eps
            The eigensolver object.
        """
        ...

    def getCyclicEPS(self) -> EPS:
        """
        Get the eigensolver object associated with the singular value solver.

        Returns
        -------
        EPS
            The eigensolver object.
        """
        ...

    def setCyclicExplicitMatrix(self, flag: bool = True) -> None:
        """
        Set if the eigensolver operator H(A) must be computed explicitly.

        Parameters
        ----------
        flag
            True if H(A) is built explicitly.
        """
        ...

    def getCyclicExplicitMatrix(self) -> bool:
        """
        Get the flag indicating if H(A) is built explicitly.

        Returns
        -------
        bool
            True if H(A) is built explicitly.
        """
        ...

    # === Lanczos-specific methods ===

    def setLanczosOneSide(self, flag: bool = True) -> None:
        """
        Set if the variant of the Lanczos method is one-sided or two-sided.

        Parameters
        ----------
        flag
            True if the method is one-sided.
        """
        ...

    def getLanczosOneSide(self) -> bool:
        """
        Get if the variant of the Lanczos method is one-sided or two-sided.

        Returns
        -------
        bool
            True if the method is one-sided.
        """
        ...

    # === TRLanczos-specific methods ===

    def setTRLanczosOneSide(self, flag: bool = True) -> None:
        """
        Set if the variant of the thick-restart Lanczos method is one-sided.

        Parameters
        ----------
        flag
            True if the method is one-sided.
        """
        ...

    def getTRLanczosOneSide(self) -> bool:
        """
        Get if the variant of thick-restart Lanczos is one-sided.

        Returns
        -------
        bool
            True if the method is one-sided.
        """
        ...

    def setTRLanczosGBidiag(self, bidiag: SVDTRLanczosGBidiag) -> None:
        """
        Set the bidiagonalization choice to use in the GSVD TRLanczos solver.

        Parameters
        ----------
        bidiag
            The bidiagonalization choice.
        """
        ...

    def getTRLanczosGBidiag(self) -> SVDTRLanczosGBidiag:
        """
        Get bidiagonalization choice used in the GSVD TRLanczos solver.

        Returns
        -------
        SVDTRLanczosGBidiag
            The bidiagonalization choice.
        """
        ...

    def setTRLanczosRestart(self, keep: float) -> None:
        """
        Set the restart parameter for the thick-restart Lanczos method.

        Parameters
        ----------
        keep
            The number of vectors to be kept at restart.
        """
        ...

    def getTRLanczosRestart(self) -> float:
        """
        Get the restart parameter used in the thick-restart Lanczos method.

        Returns
        -------
        float
            The number of vectors to be kept at restart.
        """
        ...

    def setTRLanczosLocking(self, lock: bool) -> None:
        """
        Toggle between locking and non-locking variants of the method.

        Parameters
        ----------
        lock
            True if the locking variant must be selected.
        """
        ...

    def getTRLanczosLocking(self) -> bool:
        """
        Get the locking flag used in the thick-restart Lanczos method.

        Returns
        -------
        bool
            The locking flag.
        """
        ...

    def setTRLanczosKSP(self, ksp: KSP) -> None:
        """
        Set a linear solver object associated with the SVD solver.

        Parameters
        ----------
        ksp
            The linear solver object.
        """
        ...

    def getTRLanczosKSP(self) -> KSP:
        """
        Get the linear solver object associated with the SVD solver.

        Returns
        -------
        KSP
            The linear solver object.
        """
        ...

    def setTRLanczosExplicitMatrix(self, flag: bool = True) -> None:
        """
        Set if the matrix Z=[A;B] must be built explicitly.

        Parameters
        ----------
        flag
            True if Z=[A;B] is built explicitly.
        """
        ...

    def getTRLanczosExplicitMatrix(self) -> bool:
        """
        Get the flag indicating if Z=[A;B] is built explicitly.

        Returns
        -------
        bool
            True if Z=[A;B] is built explicitly.
        """
        ...

    # --- Properties ---

    @property
    def problem_type(self) -> SVDProblemType:
        """The type of the singular value problem."""
        ...

    @problem_type.setter
    def problem_type(self, value: SVDProblemType) -> None: ...
    @property
    def transpose_mode(self) -> bool:
        """How to handle the transpose of the matrix."""
        ...

    @transpose_mode.setter
    def transpose_mode(self, value: bool) -> None: ...
    @property
    def which(self) -> SVDWhich:
        """The portion of the spectrum to be sought."""
        ...

    @which.setter
    def which(self, value: SVDWhich) -> None: ...
    @property
    def tol(self) -> float:
        """The tolerance."""
        ...

    @tol.setter
    def tol(self, value: float) -> None: ...
    @property
    def max_it(self) -> int:
        """The maximum iteration count."""
        ...

    @max_it.setter
    def max_it(self, value: int) -> None: ...
    @property
    def track_all(self) -> bool:
        """Compute the residual norm of all approximate singular triplets."""
        ...

    @track_all.setter
    def track_all(self, value: bool) -> None: ...
    @property
    def ds(self) -> DS:
        """The direct solver (DS) object associated."""
        ...

    @ds.setter
    def ds(self, value: DS) -> None: ...



cdef extern from * nogil:

    ctypedef char* SlepcNEPType "const char*"
    SlepcNEPType NEPRII
    SlepcNEPType NEPSLP
    SlepcNEPType NEPNARNOLDI
    SlepcNEPType NEPCISS
    SlepcNEPType NEPINTERPOL
    SlepcNEPType NEPNLEIGS

    ctypedef enum SlepcNEPProblemType "NEPProblemType":
        NEP_GENERAL
        NEP_RATIONAL

    ctypedef enum SlepcNEPWhich "NEPWhich":
        NEP_LARGEST_MAGNITUDE
        NEP_SMALLEST_MAGNITUDE
        NEP_LARGEST_REAL
        NEP_SMALLEST_REAL
        NEP_LARGEST_IMAGINARY
        NEP_SMALLEST_IMAGINARY
        NEP_TARGET_MAGNITUDE
        NEP_TARGET_REAL
        NEP_TARGET_IMAGINARY
        NEP_ALL
        NEP_WHICH_USER

    ctypedef enum SlepcNEPErrorType "NEPErrorType":
        NEP_ERROR_ABSOLUTE
        NEP_ERROR_RELATIVE
        NEP_ERROR_BACKWARD

    ctypedef enum SlepcNEPRefine "NEPRefine":
        NEP_REFINE_NONE
        NEP_REFINE_SIMPLE
        NEP_REFINE_MULTIPLE

    ctypedef enum SlepcNEPRefineScheme "NEPRefineScheme":
        NEP_REFINE_SCHEME_SCHUR
        NEP_REFINE_SCHEME_MBE
        NEP_REFINE_SCHEME_EXPLICIT

    ctypedef enum SlepcNEPConv "NEPConv":
        NEP_CONV_ABS
        NEP_CONV_REL
        NEP_CONV_NORM
        NEP_CONV_USER

    ctypedef enum SlepcNEPStop "NEPStop":
        NEP_STOP_BASIC
        NEP_STOP_USER

    ctypedef enum SlepcNEPConvergedReason "NEPConvergedReason":
        NEP_CONVERGED_TOL
        NEP_CONVERGED_USER
        NEP_DIVERGED_ITS
        NEP_DIVERGED_BREAKDOWN
        NEP_DIVERGED_LINEAR_SOLVE
        NEP_DIVERGED_SUBSPACE_EXHAUSTED
        NEP_CONVERGED_ITERATING

    ctypedef PetscErrorCode (*SlepcNEPFunction)(SlepcNEP,
                                     PetscScalar,
                                     PetscMat,
                                     PetscMat,
                                     void*) except PETSC_ERR_PYTHON

    ctypedef PetscErrorCode (*SlepcNEPJacobian)(SlepcNEP,
                                     PetscScalar,
                                     PetscMat,
                                     void*) except PETSC_ERR_PYTHON

    ctypedef PetscErrorCode (*SlepcNEPCtxDel)(void*)
    ctypedef PetscErrorCode (*SlepcNEPStoppingFunction)(SlepcNEP,
                                             PetscInt,
                                             PetscInt,
                                             PetscInt,
                                             PetscInt,
                                             SlepcNEPConvergedReason*,
                                             void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*SlepcNEPMonitorFunction)(SlepcNEP,
                                            PetscInt,
                                            PetscInt,
                                            PetscScalar*,
                                            PetscScalar*,
                                            PetscReal*,
                                            PetscInt,
                                            void*) except PETSC_ERR_PYTHON

    PetscErrorCode NEPCreate(MPI_Comm,SlepcNEP*)
    PetscErrorCode NEPDestroy(SlepcNEP*)
    PetscErrorCode NEPReset(SlepcNEP)
    PetscErrorCode NEPView(SlepcNEP,PetscViewer)

    PetscErrorCode NEPSetType(SlepcNEP,SlepcNEPType)
    PetscErrorCode NEPGetType(SlepcNEP,SlepcNEPType*)
    PetscErrorCode NEPSetTarget(SlepcNEP,PetscScalar)
    PetscErrorCode NEPGetTarget(SlepcNEP,PetscScalar*)
    PetscErrorCode NEPSetOptionsPrefix(SlepcNEP,char*)
    PetscErrorCode NEPGetOptionsPrefix(SlepcNEP,char*[])
    PetscErrorCode NEPSetFromOptions(SlepcNEP)
    PetscErrorCode NEPAppendOptionsPrefix(SlepcNEP,char*)
    PetscErrorCode NEPSetUp(SlepcNEP)
    PetscErrorCode NEPSolve(SlepcNEP)

    PetscErrorCode NEPSetFunction(SlepcNEP,PetscMat,PetscMat,SlepcNEPFunction,void*)
    PetscErrorCode NEPGetFunction(SlepcNEP,PetscMat*,PetscMat*,SlepcNEPFunction*,void**)
    PetscErrorCode NEPSetJacobian(SlepcNEP,PetscMat,SlepcNEPJacobian,void*)
    PetscErrorCode NEPGetJacobian(SlepcNEP,PetscMat*,SlepcNEPJacobian*,void**)
    PetscErrorCode NEPSetSplitOperator(SlepcNEP,PetscInt,PetscMat[],SlepcFN[],PetscMatStructure)
    PetscErrorCode NEPGetSplitOperatorTerm(SlepcNEP,PetscInt,PetscMat*,SlepcFN*)
    PetscErrorCode NEPGetSplitOperatorInfo(SlepcNEP,PetscInt*,PetscMatStructure*)
    PetscErrorCode NEPSetSplitPreconditioner(SlepcNEP,PetscInt,PetscMat[],PetscMatStructure)
    PetscErrorCode NEPGetSplitPreconditionerTerm(SlepcNEP,PetscInt,PetscMat*)
    PetscErrorCode NEPGetSplitPreconditionerInfo(SlepcNEP,PetscInt*,PetscMatStructure*)

    PetscErrorCode NEPSetBV(SlepcNEP,SlepcBV)
    PetscErrorCode NEPGetBV(SlepcNEP,SlepcBV*)
    PetscErrorCode NEPSetRG(SlepcNEP,SlepcRG)
    PetscErrorCode NEPGetRG(SlepcNEP,SlepcRG*)
    PetscErrorCode NEPSetDS(SlepcNEP,SlepcDS)
    PetscErrorCode NEPGetDS(SlepcNEP,SlepcDS*)
    PetscErrorCode NEPSetTolerances(SlepcNEP,PetscReal,PetscInt)
    PetscErrorCode NEPGetTolerances(SlepcNEP,PetscReal*,PetscInt*)

    PetscErrorCode NEPSetTwoSided(SlepcNEP,PetscBool)
    PetscErrorCode NEPGetTwoSided(SlepcNEP,PetscBool*)
    PetscErrorCode NEPApplyResolvent(SlepcNEP,SlepcRG,PetscScalar,PetscVec,PetscVec)

    PetscErrorCode NEPSetTrackAll(SlepcNEP,PetscBool)
    PetscErrorCode NEPGetTrackAll(SlepcNEP,PetscBool*)

    PetscErrorCode NEPSetDimensions(SlepcNEP,PetscInt,PetscInt,PetscInt)
    PetscErrorCode NEPGetDimensions(SlepcNEP,PetscInt*,PetscInt*,PetscInt*)

    PetscErrorCode NEPGetConverged(SlepcNEP,PetscInt*)
    PetscErrorCode NEPGetEigenpair(SlepcNEP,PetscInt,PetscScalar*,PetscScalar*,PetscVec,PetscVec)
    PetscErrorCode NEPGetLeftEigenvector(SlepcNEP,PetscInt,PetscVec,PetscVec)
    PetscErrorCode NEPComputeError(SlepcNEP,PetscInt,SlepcNEPErrorType,PetscReal*)
    PetscErrorCode NEPErrorView(SlepcNEP,SlepcNEPErrorType,PetscViewer)
    PetscErrorCode NEPValuesView(SlepcNEP,PetscViewer)
    PetscErrorCode NEPVectorsView(SlepcNEP,PetscViewer)
    PetscErrorCode NEPGetErrorEstimate(SlepcNEP,PetscInt,PetscReal*)

    PetscErrorCode NEPMonitorSet(SlepcNEP,SlepcNEPMonitorFunction,void*,SlepcNEPCtxDel)
    PetscErrorCode NEPMonitorCancel(SlepcNEP)
    PetscErrorCode NEPGetIterationNumber(SlepcNEP,PetscInt*)

    PetscErrorCode NEPSetInitialSpace(SlepcNEP,PetscInt,PetscVec*)
    PetscErrorCode NEPSetProblemType(SlepcNEP,SlepcNEPProblemType)
    PetscErrorCode NEPGetProblemType(SlepcNEP,SlepcNEPProblemType*)
    PetscErrorCode NEPSetWhichEigenpairs(SlepcNEP,SlepcNEPWhich)
    PetscErrorCode NEPGetWhichEigenpairs(SlepcNEP,SlepcNEPWhich*)

    PetscErrorCode NEPSetRefine(SlepcNEP,SlepcNEPRefine,PetscInt,PetscReal,PetscInt,SlepcNEPRefineScheme)
    PetscErrorCode NEPGetRefine(SlepcNEP,SlepcNEPRefine*,PetscInt*,PetscReal*,PetscInt*,SlepcNEPRefineScheme*)
    PetscErrorCode NEPRefineGetKSP(SlepcNEP,PetscKSP*)

    PetscErrorCode NEPGetConvergedReason(SlepcNEP,SlepcNEPConvergedReason*)
    PetscErrorCode NEPSetConvergenceTest(SlepcNEP,SlepcNEPConv)
    PetscErrorCode NEPGetConvergenceTest(SlepcNEP,SlepcNEPConv*)

    PetscErrorCode NEPSetStoppingTestFunction(SlepcNEP,SlepcNEPStoppingFunction,void*,SlepcNEPCtxDel)
    PetscErrorCode NEPStoppingBasic(SlepcNEP,PetscInt,PetscInt,PetscInt,PetscInt,SlepcNEPConvergedReason*,void*) except PETSC_ERR_PYTHON

    PetscErrorCode NEPRIISetLagPreconditioner(SlepcNEP,PetscInt)
    PetscErrorCode NEPRIIGetLagPreconditioner(SlepcNEP,PetscInt*)
    PetscErrorCode NEPRIISetConstCorrectionTol(SlepcNEP,PetscBool)
    PetscErrorCode NEPRIIGetConstCorrectionTol(SlepcNEP,PetscBool*)
    PetscErrorCode NEPRIISetMaximumIterations(SlepcNEP,PetscInt)
    PetscErrorCode NEPRIIGetMaximumIterations(SlepcNEP,PetscInt*)
    PetscErrorCode NEPRIISetHermitian(SlepcNEP,PetscBool)
    PetscErrorCode NEPRIIGetHermitian(SlepcNEP,PetscBool*)
    PetscErrorCode NEPRIISetDeflationThreshold(SlepcNEP,PetscReal)
    PetscErrorCode NEPRIIGetDeflationThreshold(SlepcNEP,PetscReal*)
    PetscErrorCode NEPRIISetKSP(SlepcNEP,PetscKSP)
    PetscErrorCode NEPRIIGetKSP(SlepcNEP,PetscKSP*)

    PetscErrorCode NEPSLPSetDeflationThreshold(SlepcNEP,PetscReal)
    PetscErrorCode NEPSLPGetDeflationThreshold(SlepcNEP,PetscReal*)
    PetscErrorCode NEPSLPSetEPS(SlepcNEP,SlepcEPS)
    PetscErrorCode NEPSLPGetEPS(SlepcNEP,SlepcEPS*)
    PetscErrorCode NEPSLPSetEPSLeft(SlepcNEP,SlepcEPS)
    PetscErrorCode NEPSLPGetEPSLeft(SlepcNEP,SlepcEPS*)
    PetscErrorCode NEPSLPSetKSP(SlepcNEP,PetscKSP)
    PetscErrorCode NEPSLPGetKSP(SlepcNEP,PetscKSP*)

    PetscErrorCode NEPNArnoldiSetKSP(SlepcNEP,PetscKSP)
    PetscErrorCode NEPNArnoldiGetKSP(SlepcNEP,PetscKSP*)
    PetscErrorCode NEPNArnoldiSetLagPreconditioner(SlepcNEP,PetscInt)
    PetscErrorCode NEPNArnoldiGetLagPreconditioner(SlepcNEP,PetscInt*)

    PetscErrorCode NEPInterpolSetPEP(SlepcNEP,SlepcPEP)
    PetscErrorCode NEPInterpolGetPEP(SlepcNEP,SlepcPEP*)
    PetscErrorCode NEPInterpolSetInterpolation(SlepcNEP,PetscReal,PetscInt)
    PetscErrorCode NEPInterpolGetInterpolation(SlepcNEP,PetscReal*,PetscInt*)

    PetscErrorCode NEPNLEIGSSetRestart(SlepcNEP,PetscReal)
    PetscErrorCode NEPNLEIGSGetRestart(SlepcNEP,PetscReal*)
    PetscErrorCode NEPNLEIGSSetLocking(SlepcNEP,PetscBool)
    PetscErrorCode NEPNLEIGSGetLocking(SlepcNEP,PetscBool*)
    PetscErrorCode NEPNLEIGSSetInterpolation(SlepcNEP,PetscReal,PetscInt)
    PetscErrorCode NEPNLEIGSGetInterpolation(SlepcNEP,PetscReal*,PetscInt*)
    PetscErrorCode NEPNLEIGSSetRKShifts(SlepcNEP,PetscInt,PetscScalar[])
    PetscErrorCode NEPNLEIGSGetRKShifts(SlepcNEP,PetscInt*,PetscScalar*[])
    PetscErrorCode NEPNLEIGSGetKSPs(SlepcNEP,PetscInt*,PetscKSP**)
    PetscErrorCode NEPNLEIGSSetFullBasis(SlepcNEP,PetscBool)
    PetscErrorCode NEPNLEIGSGetFullBasis(SlepcNEP,PetscBool*)
    PetscErrorCode NEPNLEIGSSetEPS(SlepcNEP,SlepcEPS)
    PetscErrorCode NEPNLEIGSGetEPS(SlepcNEP,SlepcEPS*)

    ctypedef enum SlepcNEPCISSExtraction "NEPCISSExtraction":
        NEP_CISS_EXTRACTION_RITZ
        NEP_CISS_EXTRACTION_HANKEL
        NEP_CISS_EXTRACTION_CAA

    PetscErrorCode NEPCISSSetExtraction(SlepcNEP,SlepcNEPCISSExtraction)
    PetscErrorCode NEPCISSGetExtraction(SlepcNEP,SlepcNEPCISSExtraction*)
    PetscErrorCode NEPCISSSetSizes(SlepcNEP,PetscInt,PetscInt,PetscInt,PetscInt,PetscInt,PetscBool)
    PetscErrorCode NEPCISSGetSizes(SlepcNEP,PetscInt*,PetscInt*,PetscInt*,PetscInt*,PetscInt*,PetscBool*)
    PetscErrorCode NEPCISSSetThreshold(SlepcNEP,PetscReal,PetscReal)
    PetscErrorCode NEPCISSGetThreshold(SlepcNEP,PetscReal*,PetscReal*)
    PetscErrorCode NEPCISSSetRefinement(SlepcNEP,PetscInt,PetscInt)
    PetscErrorCode NEPCISSGetRefinement(SlepcNEP,PetscInt*,PetscInt*)
    PetscErrorCode NEPCISSGetKSPs(SlepcNEP,PetscInt*,PetscKSP**)

# -----------------------------------------------------------------------------

cdef inline NEP ref_NEP(SlepcNEP nep):
    cdef NEP ob = <NEP> NEP()
    ob.nep = nep
    CHKERR( PetscINCREF(ob.obj) )
    return ob

# -----------------------------------------------------------------------------

cdef PetscErrorCode NEP_Function(
    SlepcNEP    nep,
    PetscScalar mu,
    PetscMat    A,
    PetscMat    B,
    void*       ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef NEP Nep  = ref_NEP(nep)
    cdef Mat Amat = ref_Mat(A)
    cdef Mat Bmat = ref_Mat(B)
    (function, args, kargs) = Nep.get_attr('__function__')
    retv = function(Nep, toScalar(mu), Amat, Bmat, *args, **kargs)
    cdef PetscMat Atmp = NULL, Btmp = NULL
    Atmp = A; A = Amat.mat; Amat.mat = Atmp
    Btmp = B; B = Bmat.mat; Bmat.mat = Btmp
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode NEP_Jacobian(
    SlepcNEP    nep,
    PetscScalar mu,
    PetscMat    J,
    void*       ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef NEP Nep  = ref_NEP(nep)
    cdef Mat Jmat = ref_Mat(J)
    (jacobian, args, kargs) = Nep.get_attr('__jacobian__')
    retv = jacobian(Nep, toScalar(mu), Jmat, *args, **kargs)
    cdef PetscMat Jtmp = NULL
    Jtmp = J; J = Jmat.mat; Jmat.mat = Jtmp
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode NEP_Stopping(
    SlepcNEP                nep,
    PetscInt                its,
    PetscInt                max_it,
    PetscInt                nconv,
    PetscInt                nev,
    SlepcNEPConvergedReason *r,
    void                    *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef NEP Nep = ref_NEP(nep)
    (stopping, args, kargs) = Nep.get_attr('__stopping__')
    reason = stopping(Nep, toInt(its), toInt(max_it), toInt(nconv), toInt(nev), *args, **kargs)
    if   reason is None:  r[0] = NEP_CONVERGED_ITERATING
    elif reason is False: r[0] = NEP_CONVERGED_ITERATING
    elif reason is True:  r[0] = NEP_CONVERGED_USER
    else:                 r[0] = reason
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode NEP_Monitor(
    SlepcNEP    nep,
    PetscInt    its,
    PetscInt    nconv,
    PetscScalar *eigr,
    PetscScalar *eigi,
    PetscReal   *errest,
    PetscInt    nest,
    void        *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef NEP Nep = ref_NEP(nep)
    cdef object monitorlist = Nep.get_attr('__monitor__')
    if monitorlist is None: return PETSC_SUCCESS
    cdef object eig = [toComplex(eigr[i], eigi[i]) for i in range(nest)]
    cdef object err = [toReal(errest[i]) for i in range(nest)]
    for (monitor, args, kargs) in monitorlist:
        monitor(Nep, toInt(its), toInt(nconv), eig, err, *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------



#cython: language_level=3str
#cython: embedsignature=True
#cython: embedsignature.format=python
#cython: annotation_typing=False
#cython: cdivision=True
#cython: auto_pickle=False
#cython: always_allow_keywords=True
#cython: allow_none_for_extension_args=False
#cython: autotestdict=False
#cython: warn.multiple_declarators=False
#cython: optimize.use_switch=False
#cython: binding=True

from __future__ import annotations
cimport cython

include "SLEPc/SLEPc.pyx"



cdef extern from * nogil:

    ctypedef char* SlepcSVDType "const char*"
    SlepcSVDType SVDCROSS
    SlepcSVDType SVDCYCLIC
    SlepcSVDType SVDLAPACK
    SlepcSVDType SVDLANCZOS
    SlepcSVDType SVDTRLANCZOS
    SlepcSVDType SVDRANDOMIZED
    SlepcSVDType SVDSCALAPACK
    SlepcSVDType SVDKSVD
    SlepcSVDType SVDELEMENTAL
    SlepcSVDType SVDPRIMME

    ctypedef enum SlepcSVDProblemType "SVDProblemType":
        SVD_STANDARD
        SVD_GENERALIZED
        SVD_HYPERBOLIC

    ctypedef enum SlepcSVDWhich "SVDWhich":
        SVD_LARGEST
        SVD_SMALLEST

    ctypedef enum SlepcSVDErrorType "SVDErrorType":
        SVD_ERROR_ABSOLUTE
        SVD_ERROR_RELATIVE
        SVD_ERROR_NORM

    ctypedef enum SlepcSVDConv "SVDConv":
        SVD_CONV_ABS
        SVD_CONV_REL
        SVD_CONV_NORM
        SVD_CONV_MAXIT
        SVD_CONV_USER

    ctypedef enum SlepcSVDStop "SVDStop":
        SVD_STOP_BASIC
        SVD_STOP_USER
        SVD_STOP_THRESHOLD

    ctypedef enum SlepcSVDConvergedReason "SVDConvergedReason":
        SVD_CONVERGED_TOL
        SVD_CONVERGED_USER
        SVD_CONVERGED_MAXIT
        SVD_DIVERGED_ITS
        SVD_DIVERGED_BREAKDOWN
        SVD_DIVERGED_SYMMETRY_LOST
        SVD_CONVERGED_ITERATING

    ctypedef PetscErrorCode (*SlepcSVDCtxDel)(void*)
    ctypedef PetscErrorCode (*SlepcSVDStoppingFunction)(SlepcSVD,
                                             PetscInt,
                                             PetscInt,
                                             PetscInt,
                                             PetscInt,
                                             SlepcSVDConvergedReason*,
                                             void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*SlepcSVDMonitorFunction)(SlepcSVD,
                                            PetscInt,
                                            PetscInt,
                                            PetscReal*,
                                            PetscReal*,
                                            PetscInt,
                                            void*) except PETSC_ERR_PYTHON

    PetscErrorCode SVDCreate(MPI_Comm,SlepcSVD*)
    PetscErrorCode SVDView(SlepcSVD,PetscViewer)
    PetscErrorCode SVDDestroy(SlepcSVD*)
    PetscErrorCode SVDReset(SlepcSVD)
    PetscErrorCode SVDSetType(SlepcSVD,SlepcSVDType)
    PetscErrorCode SVDGetType(SlepcSVD,SlepcSVDType*)
    PetscErrorCode SVDSetOptionsPrefix(SlepcSVD,char[])
    PetscErrorCode SVDAppendOptionsPrefix(SlepcSVD,char[])
    PetscErrorCode SVDGetOptionsPrefix(SlepcSVD,char*[])
    PetscErrorCode SVDSetFromOptions(SlepcSVD)

    PetscErrorCode SVDSetProblemType(SlepcSVD,SlepcSVDProblemType)
    PetscErrorCode SVDGetProblemType(SlepcSVD,SlepcSVDProblemType*)
    PetscErrorCode SVDIsGeneralized(SlepcSVD,PetscBool*)
    PetscErrorCode SVDIsHyperbolic(SlepcSVD,PetscBool*)
    PetscErrorCode SVDSetBV(SlepcSVD,SlepcBV,SlepcBV)
    PetscErrorCode SVDGetBV(SlepcSVD,SlepcBV*,SlepcBV*)
    PetscErrorCode SVDSetDS(SlepcSVD,SlepcDS)
    PetscErrorCode SVDGetDS(SlepcSVD,SlepcDS*)

    PetscErrorCode SVDSetOperators(SlepcSVD,PetscMat,PetscMat)
    PetscErrorCode SVDGetOperators(SlepcSVD,PetscMat*,PetscMat*)
    PetscErrorCode SVDSetSignature(SlepcSVD,PetscVec)
    PetscErrorCode SVDGetSignature(SlepcSVD,PetscVec)

    PetscErrorCode SVDSetInitialSpaces(SlepcSVD,PetscInt,PetscVec*,PetscInt,PetscVec*)

    PetscErrorCode SVDSetImplicitTranspose(SlepcSVD,PetscBool)
    PetscErrorCode SVDGetImplicitTranspose(SlepcSVD,PetscBool*)
    PetscErrorCode SVDSetDimensions(SlepcSVD,PetscInt,PetscInt,PetscInt)
    PetscErrorCode SVDGetDimensions(SlepcSVD,PetscInt*,PetscInt*,PetscInt*)
    PetscErrorCode SVDSetTolerances(SlepcSVD,PetscReal,PetscInt)
    PetscErrorCode SVDGetTolerances(SlepcSVD,PetscReal*,PetscInt*)
    PetscErrorCode SVDSetWhichSingularTriplets(SlepcSVD,SlepcSVDWhich)
    PetscErrorCode SVDGetWhichSingularTriplets(SlepcSVD,SlepcSVDWhich*)
    PetscErrorCode SVDSetThreshold(SlepcSVD,PetscReal,PetscBool)
    PetscErrorCode SVDGetThreshold(SlepcSVD,PetscReal*,PetscBool*)
    PetscErrorCode SVDSetConvergenceTest(SlepcSVD,SlepcSVDConv)
    PetscErrorCode SVDGetConvergenceTest(SlepcSVD,SlepcSVDConv*)

    PetscErrorCode SVDMonitorSet(SlepcSVD,SlepcSVDMonitorFunction,void*,SlepcSVDCtxDel)
    PetscErrorCode SVDMonitorCancel(SlepcSVD)

    PetscErrorCode SVDSetStoppingTestFunction(SlepcSVD,SlepcSVDStoppingFunction,void*,SlepcSVDCtxDel)
    PetscErrorCode SVDStoppingBasic(SlepcSVD,PetscInt,PetscInt,PetscInt,PetscInt,SlepcSVDConvergedReason*,void*) except PETSC_ERR_PYTHON

    PetscErrorCode SVDSetTrackAll(SlepcSVD,PetscBool)
    PetscErrorCode SVDGetTrackAll(SlepcSVD,PetscBool*)

    PetscErrorCode SVDSetUp(SlepcSVD)
    PetscErrorCode SVDSolve(SlepcSVD)
    PetscErrorCode SVDGetIterationNumber(SlepcSVD,PetscInt*)
    PetscErrorCode SVDGetConvergedReason(SlepcSVD,SlepcSVDConvergedReason*)
    PetscErrorCode SVDGetConverged(SlepcSVD,PetscInt*)
    PetscErrorCode SVDGetSingularTriplet(SlepcSVD,PetscInt,PetscReal*,PetscVec,PetscVec)
    PetscErrorCode SVDComputeError(SlepcSVD,PetscInt,SlepcSVDErrorType,PetscReal*)
    PetscErrorCode SVDErrorView(SlepcSVD,SlepcSVDErrorType,PetscViewer)
    PetscErrorCode SVDValuesView(SlepcSVD,PetscViewer)
    PetscErrorCode SVDVectorsView(SlepcSVD,PetscViewer)

    PetscErrorCode SVDCrossSetExplicitMatrix(SlepcSVD,PetscBool)
    PetscErrorCode SVDCrossGetExplicitMatrix(SlepcSVD,PetscBool*)
    PetscErrorCode SVDCrossSetEPS(SlepcSVD,SlepcEPS)
    PetscErrorCode SVDCrossGetEPS(SlepcSVD,SlepcEPS*)

    PetscErrorCode SVDCyclicSetExplicitMatrix(SlepcSVD,PetscBool)
    PetscErrorCode SVDCyclicGetExplicitMatrix(SlepcSVD,PetscBool*)
    PetscErrorCode SVDCyclicSetEPS(SlepcSVD,SlepcEPS)
    PetscErrorCode SVDCyclicGetEPS(SlepcSVD,SlepcEPS*)

    PetscErrorCode SVDLanczosSetOneSide(SlepcSVD,PetscBool)
    PetscErrorCode SVDLanczosGetOneSide(SlepcSVD,PetscBool*)

    PetscErrorCode SVDTRLanczosSetOneSide(SlepcSVD,PetscBool)
    PetscErrorCode SVDTRLanczosGetOneSide(SlepcSVD,PetscBool*)
    PetscErrorCode SVDTRLanczosSetGBidiag(SlepcSVD,SlepcSVDTRLanczosGBidiag)
    PetscErrorCode SVDTRLanczosGetGBidiag(SlepcSVD,SlepcSVDTRLanczosGBidiag*)
    PetscErrorCode SVDTRLanczosSetKSP(SlepcSVD,PetscKSP)
    PetscErrorCode SVDTRLanczosGetKSP(SlepcSVD,PetscKSP*)
    PetscErrorCode SVDTRLanczosSetRestart(SlepcSVD,PetscReal)
    PetscErrorCode SVDTRLanczosGetRestart(SlepcSVD,PetscReal*)
    PetscErrorCode SVDTRLanczosSetLocking(SlepcSVD,PetscBool)
    PetscErrorCode SVDTRLanczosGetLocking(SlepcSVD,PetscBool*)
    PetscErrorCode SVDTRLanczosSetExplicitMatrix(SlepcSVD,PetscBool)
    PetscErrorCode SVDTRLanczosGetExplicitMatrix(SlepcSVD,PetscBool*)

    ctypedef enum SlepcSVDTRLanczosGBidiag "SVDTRLanczosGBidiag":
        SVD_TRLANCZOS_GBIDIAG_SINGLE
        SVD_TRLANCZOS_GBIDIAG_UPPER
        SVD_TRLANCZOS_GBIDIAG_LOWER

# -----------------------------------------------------------------------------

cdef inline SVD ref_SVD(SlepcSVD svd):
    cdef SVD ob = <SVD> SVD()
    ob.svd = svd
    CHKERR( PetscINCREF(ob.obj) )
    return ob

# -----------------------------------------------------------------------------

cdef PetscErrorCode SVD_Stopping(
    SlepcSVD                svd,
    PetscInt                its,
    PetscInt                max_it,
    PetscInt                nconv,
    PetscInt                nev,
    SlepcSVDConvergedReason *r,
    void                    *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef SVD Svd = ref_SVD(svd)
    (stopping, args, kargs) = Svd.get_attr('__stopping__')
    reason = stopping(Svd, toInt(its), toInt(max_it), toInt(nconv), toInt(nev), *args, **kargs)
    if   reason is None:  r[0] = SVD_CONVERGED_ITERATING
    elif reason is False: r[0] = SVD_CONVERGED_ITERATING
    elif reason is True:  r[0] = SVD_CONVERGED_USER
    else:                 r[0] = reason
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode SVD_Monitor(
    SlepcSVD    svd,
    PetscInt    its,
    PetscInt    nconv,
    PetscReal   *sigma,
    PetscReal   *errest,
    PetscInt    nest,
    void        *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef SVD Svd = ref_SVD(svd)
    cdef object monitorlist = Svd.get_attr('__monitor__')
    if monitorlist is None: return PETSC_SUCCESS
    cdef object eig = [toReal(sigma[i]) for i in range(nest)]
    cdef object err = [toReal(errest[i]) for i in range(nest)]
    for (monitor, args, kargs) in monitorlist:
        monitor(Svd, toInt(its), toInt(nconv), eig, err, *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------



cdef extern from * nogil:

    ctypedef char* SlepcEPSType "const char*"
    SlepcEPSType EPSPOWER
    SlepcEPSType EPSSUBSPACE
    SlepcEPSType EPSARNOLDI
    SlepcEPSType EPSLANCZOS
    SlepcEPSType EPSKRYLOVSCHUR
    SlepcEPSType EPSGD
    SlepcEPSType EPSJD
    SlepcEPSType EPSRQCG
    SlepcEPSType EPSLOBPCG
    SlepcEPSType EPSCISS
    SlepcEPSType EPSLYAPII
    SlepcEPSType EPSLAPACK
    SlepcEPSType EPSARPACK
    SlepcEPSType EPSBLOPEX
    SlepcEPSType EPSPRIMME
    SlepcEPSType EPSFEAST
    SlepcEPSType EPSSCALAPACK
    SlepcEPSType EPSELPA
    SlepcEPSType EPSELEMENTAL
    SlepcEPSType EPSEVSL
    SlepcEPSType EPSCHASE

    ctypedef enum SlepcEPSProblemType "EPSProblemType":
        EPS_HEP
        EPS_GHEP
        EPS_NHEP
        EPS_GNHEP
        EPS_PGNHEP
        EPS_GHIEP
        EPS_BSE
        EPS_HAMILT

    ctypedef enum SlepcEPSExtraction "EPSExtraction":
        EPS_RITZ
        EPS_HARMONIC
        EPS_HARMONIC_RELATIVE
        EPS_HARMONIC_RIGHT
        EPS_HARMONIC_LARGEST
        EPS_REFINED
        EPS_REFINED_HARMONIC

    ctypedef enum SlepcEPSWhich "EPSWhich":
        EPS_LARGEST_MAGNITUDE
        EPS_LARGEST_REAL
        EPS_LARGEST_IMAGINARY
        EPS_SMALLEST_MAGNITUDE
        EPS_SMALLEST_REAL
        EPS_SMALLEST_IMAGINARY
        EPS_TARGET_MAGNITUDE
        EPS_TARGET_REAL
        EPS_TARGET_IMAGINARY
        EPS_ALL
        EPS_WHICH_USER

    ctypedef enum SlepcEPSBalance "EPSBalance":
        EPS_BALANCE_NONE
        EPS_BALANCE_ONESIDE
        EPS_BALANCE_TWOSIDE
        EPS_BALANCE_USER

    ctypedef enum SlepcEPSErrorType "EPSErrorType":
        EPS_ERROR_ABSOLUTE
        EPS_ERROR_RELATIVE
        EPS_ERROR_BACKWARD

    ctypedef enum SlepcEPSConv "EPSConv":
        EPS_CONV_ABS
        EPS_CONV_REL
        EPS_CONV_NORM
        EPS_CONV_USER

    ctypedef enum SlepcEPSStop "EPSStop":
        EPS_STOP_BASIC
        EPS_STOP_USER
        EPS_STOP_THRESHOLD

    ctypedef enum SlepcEPSConvergedReason "EPSConvergedReason":
        EPS_CONVERGED_TOL
        EPS_CONVERGED_USER
        EPS_DIVERGED_ITS
        EPS_DIVERGED_BREAKDOWN
        EPS_DIVERGED_SYMMETRY_LOST
        EPS_CONVERGED_ITERATING

    ctypedef PetscErrorCode (*SlepcEPSCtxDel)(void*)
    ctypedef PetscErrorCode (*SlepcEPSStoppingFunction)(SlepcEPS,
                                             PetscInt,
                                             PetscInt,
                                             PetscInt,
                                             PetscInt,
                                             SlepcEPSConvergedReason*,
                                             void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*SlepcEPSMonitorFunction)(SlepcEPS,
                                            PetscInt,
                                            PetscInt,
                                            PetscScalar*,
                                            PetscScalar*,
                                            PetscReal*,
                                            PetscInt,
                                            void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*SlepcEPSArbitraryFunction)(PetscScalar,
                                             PetscScalar,
                                             PetscVec,
                                             PetscVec,
                                             PetscScalar*,
                                             PetscScalar*,
                                             void*) except PETSC_ERR_PYTHON
    ctypedef PetscErrorCode (*SlepcEPSComparisonFunction)(PetscScalar,
                                               PetscScalar,
                                               PetscScalar,
                                               PetscScalar,
                                               PetscInt*,
                                               void*) except PETSC_ERR_PYTHON

    PetscErrorCode EPSView(SlepcEPS,PetscViewer)
    PetscErrorCode EPSDestroy(SlepcEPS*)
    PetscErrorCode EPSReset(SlepcEPS)
    PetscErrorCode EPSCreate(MPI_Comm,SlepcEPS*)
    PetscErrorCode EPSSetType(SlepcEPS,SlepcEPSType)
    PetscErrorCode EPSGetType(SlepcEPS,SlepcEPSType*)
    PetscErrorCode EPSSetOptionsPrefix(SlepcEPS,char[])
    PetscErrorCode EPSAppendOptionsPrefix(SlepcEPS,char [])
    PetscErrorCode EPSGetOptionsPrefix(SlepcEPS,char*[])
    PetscErrorCode EPSSetFromOptions(SlepcEPS)

    PetscErrorCode EPSSetProblemType(SlepcEPS,SlepcEPSProblemType)
    PetscErrorCode EPSGetProblemType(SlepcEPS,SlepcEPSProblemType*)
    PetscErrorCode EPSIsGeneralized(SlepcEPS,PetscBool*)
    PetscErrorCode EPSIsHermitian(SlepcEPS,PetscBool*)
    PetscErrorCode EPSIsPositive(SlepcEPS,PetscBool*)
    PetscErrorCode EPSIsStructured(SlepcEPS,PetscBool*)
    PetscErrorCode EPSSetExtraction(SlepcEPS,SlepcEPSExtraction)
    PetscErrorCode EPSGetExtraction(SlepcEPS,SlepcEPSExtraction*)
    PetscErrorCode EPSSetBalance(SlepcEPS,SlepcEPSBalance,PetscInt,PetscReal)
    PetscErrorCode EPSGetBalance(SlepcEPS,SlepcEPSBalance*,PetscInt*,PetscReal*)
    PetscErrorCode EPSSetWhichEigenpairs(SlepcEPS,SlepcEPSWhich)
    PetscErrorCode EPSGetWhichEigenpairs(SlepcEPS,SlepcEPSWhich*)
    PetscErrorCode EPSSetThreshold(SlepcEPS,PetscReal,PetscBool)
    PetscErrorCode EPSGetThreshold(SlepcEPS,PetscReal*,PetscBool*)
    PetscErrorCode EPSSetTarget(SlepcEPS,PetscScalar)
    PetscErrorCode EPSGetTarget(SlepcEPS,PetscScalar*)
    PetscErrorCode EPSSetInterval(SlepcEPS,PetscReal,PetscReal)
    PetscErrorCode EPSGetInterval(SlepcEPS,PetscReal*,PetscReal*)
    PetscErrorCode EPSSetTolerances(SlepcEPS,PetscReal,PetscInt)
    PetscErrorCode EPSGetTolerances(SlepcEPS,PetscReal*,PetscInt*)
    PetscErrorCode EPSSetDimensions(SlepcEPS,PetscInt,PetscInt,PetscInt)
    PetscErrorCode EPSGetDimensions(SlepcEPS,PetscInt*,PetscInt*,PetscInt*)

    PetscErrorCode EPSSetBV(SlepcEPS,SlepcBV)
    PetscErrorCode EPSGetBV(SlepcEPS,SlepcBV*)
    PetscErrorCode EPSSetDS(SlepcEPS,SlepcDS)
    PetscErrorCode EPSGetDS(SlepcEPS,SlepcDS*)
    PetscErrorCode EPSSetST(SlepcEPS,SlepcST)
    PetscErrorCode EPSGetST(SlepcEPS,SlepcST*)
    PetscErrorCode EPSSetRG(SlepcEPS,SlepcRG)
    PetscErrorCode EPSGetRG(SlepcEPS,SlepcRG*)

    PetscErrorCode EPSSetOperators(SlepcEPS,PetscMat,PetscMat)
    PetscErrorCode EPSGetOperators(SlepcEPS,PetscMat*,PetscMat*)

    PetscErrorCode EPSSetTwoSided(SlepcEPS,PetscBool)
    PetscErrorCode EPSGetTwoSided(SlepcEPS,PetscBool*)
    PetscErrorCode EPSSetPurify(SlepcEPS,PetscBool)
    PetscErrorCode EPSGetPurify(SlepcEPS,PetscBool*)

    PetscErrorCode EPSSetConvergenceTest(SlepcEPS,SlepcEPSConv)
    PetscErrorCode EPSGetConvergenceTest(SlepcEPS,SlepcEPSConv*)

    PetscErrorCode EPSSetTrueResidual(SlepcEPS,PetscBool)
    PetscErrorCode EPSGetTrueResidual(SlepcEPS,PetscBool*)

    PetscErrorCode EPSSetTrackAll(SlepcEPS,PetscBool)
    PetscErrorCode EPSGetTrackAll(SlepcEPS,PetscBool*)

    PetscErrorCode EPSSetDeflationSpace(SlepcEPS,PetscInt,PetscVec*)
    PetscErrorCode EPSSetInitialSpace(SlepcEPS,PetscInt,PetscVec*)
    PetscErrorCode EPSSetLeftInitialSpace(SlepcEPS,PetscInt,PetscVec*)

    PetscErrorCode EPSSetUp(SlepcEPS)
    PetscErrorCode EPSSolve(SlepcEPS)

    PetscErrorCode EPSGetIterationNumber(SlepcEPS,PetscInt*)
    PetscErrorCode EPSGetConvergedReason(SlepcEPS,SlepcEPSConvergedReason*)
    PetscErrorCode EPSGetConverged(SlepcEPS,PetscInt*)
    PetscErrorCode EPSGetEigenvalue(SlepcEPS,PetscInt,PetscScalar*,PetscScalar*)
    PetscErrorCode EPSGetEigenvector(SlepcEPS,PetscInt,PetscVec,PetscVec)
    PetscErrorCode EPSGetLeftEigenvector(SlepcEPS,PetscInt,PetscVec,PetscVec)
    PetscErrorCode EPSGetEigenpair(SlepcEPS,PetscInt,PetscScalar*,PetscScalar*,PetscVec,PetscVec)
    PetscErrorCode EPSGetInvariantSubspace(SlepcEPS,PetscVec*)

    PetscErrorCode EPSSetStoppingTestFunction(SlepcEPS,SlepcEPSStoppingFunction,void*,SlepcEPSCtxDel)
    PetscErrorCode EPSStoppingBasic(SlepcEPS,PetscInt,PetscInt,PetscInt,PetscInt,SlepcEPSConvergedReason*,void*) except PETSC_ERR_PYTHON

    PetscErrorCode EPSSetArbitrarySelection(SlepcEPS,SlepcEPSArbitraryFunction,void*);
    PetscErrorCode EPSSetEigenvalueComparison(SlepcEPS,SlepcEPSComparisonFunction,void*);

    PetscErrorCode EPSGetErrorEstimate(SlepcEPS,PetscInt,PetscReal*)
    PetscErrorCode EPSComputeError(SlepcEPS,PetscInt,SlepcEPSErrorType,PetscReal*)
    PetscErrorCode EPSErrorView(SlepcEPS,SlepcEPSErrorType,PetscViewer)
    PetscErrorCode EPSValuesView(SlepcEPS,PetscViewer)
    PetscErrorCode EPSVectorsView(SlepcEPS,PetscViewer)

    PetscErrorCode EPSMonitorSet(SlepcEPS,SlepcEPSMonitorFunction,void*,SlepcEPSCtxDel)
    PetscErrorCode EPSMonitorCancel(SlepcEPS)

    ctypedef enum SlepcEPSPowerShiftType "EPSPowerShiftType":
        EPS_POWER_SHIFT_CONSTANT
        EPS_POWER_SHIFT_RAYLEIGH
        EPS_POWER_SHIFT_WILKINSON
    PetscErrorCode EPSPowerSetShiftType(SlepcEPS,SlepcEPSPowerShiftType)
    PetscErrorCode EPSPowerGetShiftType(SlepcEPS,SlepcEPSPowerShiftType*)

    PetscErrorCode EPSArnoldiSetDelayed(SlepcEPS,PetscBool)
    PetscErrorCode EPSArnoldiGetDelayed(SlepcEPS,PetscBool*)

    ctypedef enum SlepcEPSKrylovSchurBSEType "EPSKrylovSchurBSEType":
        EPS_KRYLOVSCHUR_BSE_SHAO
        EPS_KRYLOVSCHUR_BSE_GRUNING
        EPS_KRYLOVSCHUR_BSE_PROJECTEDBSE
    PetscErrorCode EPSKrylovSchurSetBSEType(SlepcEPS,SlepcEPSKrylovSchurBSEType)
    PetscErrorCode EPSKrylovSchurGetBSEType(SlepcEPS,SlepcEPSKrylovSchurBSEType*)
    PetscErrorCode EPSKrylovSchurSetRestart(SlepcEPS,PetscReal)
    PetscErrorCode EPSKrylovSchurGetRestart(SlepcEPS,PetscReal*)
    PetscErrorCode EPSKrylovSchurSetLocking(SlepcEPS,PetscBool)
    PetscErrorCode EPSKrylovSchurGetLocking(SlepcEPS,PetscBool*)
    PetscErrorCode EPSKrylovSchurSetPartitions(SlepcEPS,PetscInt)
    PetscErrorCode EPSKrylovSchurGetPartitions(SlepcEPS,PetscInt*)
    PetscErrorCode EPSKrylovSchurSetDetectZeros(SlepcEPS,PetscBool)
    PetscErrorCode EPSKrylovSchurGetDetectZeros(SlepcEPS,PetscBool*)
    PetscErrorCode EPSKrylovSchurSetDimensions(SlepcEPS,PetscInt,PetscInt,PetscInt)
    PetscErrorCode EPSKrylovSchurGetDimensions(SlepcEPS,PetscInt*,PetscInt*,PetscInt*)
    PetscErrorCode EPSKrylovSchurGetSubcommInfo(SlepcEPS,PetscInt*,PetscInt*,PetscVec*)
    PetscErrorCode EPSKrylovSchurGetSubcommPairs(SlepcEPS,PetscInt,PetscScalar*,PetscVec)
    PetscErrorCode EPSKrylovSchurGetSubcommMats(SlepcEPS,PetscMat*,PetscMat*)
    PetscErrorCode EPSKrylovSchurUpdateSubcommMats(SlepcEPS,PetscScalar,PetscScalar,PetscMat,PetscScalar,PetscScalar,PetscMat,PetscMatStructure,PetscBool)
    PetscErrorCode EPSKrylovSchurSetSubintervals(SlepcEPS,PetscReal*)
    PetscErrorCode EPSKrylovSchurGetSubintervals(SlepcEPS,PetscReal**)
    PetscErrorCode EPSKrylovSchurGetInertias(SlepcEPS,PetscInt*,PetscReal**,PetscInt**)
    PetscErrorCode EPSKrylovSchurGetKSP(SlepcEPS,PetscKSP*)

    ctypedef enum SlepcEPSLanczosReorthogType "EPSLanczosReorthogType":
        EPS_LANCZOS_REORTHOG_LOCAL
        EPS_LANCZOS_REORTHOG_FULL
        EPS_LANCZOS_REORTHOG_SELECTIVE
        EPS_LANCZOS_REORTHOG_PERIODIC
        EPS_LANCZOS_REORTHOG_PARTIAL
        EPS_LANCZOS_REORTHOG_DELAYED
    PetscErrorCode EPSLanczosSetReorthog(SlepcEPS,SlepcEPSLanczosReorthogType)
    PetscErrorCode EPSLanczosGetReorthog(SlepcEPS,SlepcEPSLanczosReorthogType*)

    PetscErrorCode EPSGDSetKrylovStart(SlepcEPS,PetscBool)
    PetscErrorCode EPSGDGetKrylovStart(SlepcEPS,PetscBool*)
    PetscErrorCode EPSGDSetBlockSize(SlepcEPS,PetscInt)
    PetscErrorCode EPSGDGetBlockSize(SlepcEPS,PetscInt*)
    PetscErrorCode EPSGDSetRestart(SlepcEPS,PetscInt,PetscInt)
    PetscErrorCode EPSGDGetRestart(SlepcEPS,PetscInt*,PetscInt*)
    PetscErrorCode EPSGDSetInitialSize(SlepcEPS,PetscInt)
    PetscErrorCode EPSGDGetInitialSize(SlepcEPS,PetscInt*)
    PetscErrorCode EPSGDSetBOrth(SlepcEPS,PetscBool)
    PetscErrorCode EPSGDGetBOrth(SlepcEPS,PetscBool*)
    PetscErrorCode EPSGDSetDoubleExpansion(SlepcEPS,PetscBool)
    PetscErrorCode EPSGDGetDoubleExpansion(SlepcEPS,PetscBool*)

    PetscErrorCode EPSJDSetKrylovStart(SlepcEPS,PetscBool)
    PetscErrorCode EPSJDGetKrylovStart(SlepcEPS,PetscBool*)
    PetscErrorCode EPSJDSetBlockSize(SlepcEPS,PetscInt)
    PetscErrorCode EPSJDGetBlockSize(SlepcEPS,PetscInt*)
    PetscErrorCode EPSJDSetRestart(SlepcEPS,PetscInt,PetscInt)
    PetscErrorCode EPSJDGetRestart(SlepcEPS,PetscInt*,PetscInt*)
    PetscErrorCode EPSJDSetInitialSize(SlepcEPS,PetscInt)
    PetscErrorCode EPSJDGetInitialSize(SlepcEPS,PetscInt*)
    PetscErrorCode EPSJDSetFix(SlepcEPS,PetscReal)
    PetscErrorCode EPSJDGetFix(SlepcEPS,PetscReal*)
    PetscErrorCode EPSJDSetConstCorrectionTol(SlepcEPS,PetscBool)
    PetscErrorCode EPSJDGetConstCorrectionTol(SlepcEPS,PetscBool*)
    PetscErrorCode EPSJDSetBOrth(SlepcEPS,PetscBool)
    PetscErrorCode EPSJDGetBOrth(SlepcEPS,PetscBool*)

    PetscErrorCode EPSRQCGSetReset(SlepcEPS,PetscInt)
    PetscErrorCode EPSRQCGGetReset(SlepcEPS,PetscInt*)

    PetscErrorCode EPSLOBPCGSetBlockSize(SlepcEPS,PetscInt)
    PetscErrorCode EPSLOBPCGGetBlockSize(SlepcEPS,PetscInt*)
    PetscErrorCode EPSLOBPCGSetRestart(SlepcEPS,PetscReal)
    PetscErrorCode EPSLOBPCGGetRestart(SlepcEPS,PetscReal*)
    PetscErrorCode EPSLOBPCGSetLocking(SlepcEPS,PetscBool)
    PetscErrorCode EPSLOBPCGGetLocking(SlepcEPS,PetscBool*)

    PetscErrorCode EPSLyapIISetRanks(SlepcEPS,PetscInt,PetscInt)
    PetscErrorCode EPSLyapIIGetRanks(SlepcEPS,PetscInt*,PetscInt*)

    ctypedef enum SlepcEPSCISSQuadRule "EPSCISSQuadRule":
        EPS_CISS_QUADRULE_TRAPEZOIDAL
        EPS_CISS_QUADRULE_CHEBYSHEV

    ctypedef enum SlepcEPSCISSExtraction "EPSCISSExtraction":
        EPS_CISS_EXTRACTION_RITZ
        EPS_CISS_EXTRACTION_HANKEL

    PetscErrorCode EPSCISSSetExtraction(SlepcEPS,SlepcEPSCISSExtraction)
    PetscErrorCode EPSCISSGetExtraction(SlepcEPS,SlepcEPSCISSExtraction*)
    PetscErrorCode EPSCISSSetQuadRule(SlepcEPS,SlepcEPSCISSQuadRule)
    PetscErrorCode EPSCISSGetQuadRule(SlepcEPS,SlepcEPSCISSQuadRule*)
    PetscErrorCode EPSCISSSetSizes(SlepcEPS,PetscInt,PetscInt,PetscInt,PetscInt,PetscInt,PetscBool)
    PetscErrorCode EPSCISSGetSizes(SlepcEPS,PetscInt*,PetscInt*,PetscInt*,PetscInt*,PetscInt*,PetscBool*)
    PetscErrorCode EPSCISSSetThreshold(SlepcEPS,PetscReal,PetscReal)
    PetscErrorCode EPSCISSGetThreshold(SlepcEPS,PetscReal*,PetscReal*)
    PetscErrorCode EPSCISSSetRefinement(SlepcEPS,PetscInt,PetscInt)
    PetscErrorCode EPSCISSGetRefinement(SlepcEPS,PetscInt*,PetscInt*)
    PetscErrorCode EPSCISSSetUseST(SlepcEPS,PetscBool)
    PetscErrorCode EPSCISSGetUseST(SlepcEPS,PetscBool*)
    PetscErrorCode EPSCISSGetKSPs(SlepcEPS,PetscInt*,PetscKSP**)

cdef extern from * nogil:
    PetscErrorCode VecDuplicate(PetscVec,PetscVec*)
    PetscErrorCode MatCreateVecs(PetscMat,PetscVec*,PetscVec*)

# -----------------------------------------------------------------------------

cdef inline EPS ref_EPS(SlepcEPS eps):
    cdef EPS ob = <EPS> EPS()
    ob.eps = eps
    CHKERR( PetscINCREF(ob.obj) )
    return ob

# -----------------------------------------------------------------------------

cdef PetscErrorCode EPS_Stopping(
    SlepcEPS                eps,
    PetscInt                its,
    PetscInt                max_it,
    PetscInt                nconv,
    PetscInt                nev,
    SlepcEPSConvergedReason *r,
    void                    *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef EPS Eps = ref_EPS(eps)
    (stopping, args, kargs) = Eps.get_attr('__stopping__')
    reason = stopping(Eps, toInt(its), toInt(max_it), toInt(nconv), toInt(nev), *args, **kargs)
    if   reason is None:  r[0] = EPS_CONVERGED_ITERATING
    elif reason is False: r[0] = EPS_CONVERGED_ITERATING
    elif reason is True:  r[0] = EPS_CONVERGED_USER
    else:                 r[0] = reason
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode EPS_Arbitrary(
    PetscScalar  er,
    PetscScalar  ei,
    PetscVec     xr,
    PetscVec     xi,
    PetscScalar* rr,
    PetscScalar* ri,
    void         *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    (arbitrary, args, kargs) = <object>ctx
    cdef Vec Vr = ref_Vec(xr)
    cdef Vec Vi = ref_Vec(xi)
    r = arbitrary(toComplex(er, ei), Vr, Vi, args, **kargs)
    if sizeof(PetscScalar) == sizeof(PetscReal):
        rr[0] = asComplexReal(r)
        ri[0] = asComplexImag(r)
    else:
        rr[0] = asScalar(r)
        ri[0] = 0.0
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode EPS_Comparison(
    PetscScalar  ar,
    PetscScalar  ai,
    PetscScalar  br,
    PetscScalar  bi,
    PetscInt*    res,
    void         *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    (comparison, args, kargs) = <object>ctx
    r = comparison(toComplex(ar, ai), toComplex(br, bi), args, **kargs)
    res[0] = asInt(r)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------

cdef PetscErrorCode EPS_Monitor(
    SlepcEPS    eps,
    PetscInt    its,
    PetscInt    nconv,
    PetscScalar *eigr,
    PetscScalar *eigi,
    PetscReal   *errest,
    PetscInt    nest,
    void        *ctx,
    ) except PETSC_ERR_PYTHON with gil:
    cdef EPS Eps = ref_EPS(eps)
    cdef object monitorlist = Eps.get_attr('__monitor__')
    if monitorlist is None: return PETSC_SUCCESS
    cdef object eig = [toComplex(eigr[i], eigi[i]) for i in range(nest)]
    cdef object err = [toReal(errest[i]) for i in range(nest)]
    for (monitor, args, kargs) in monitorlist:
        monitor(Eps, toInt(its), toInt(nconv), eig, err, *args, **kargs)
    return PETSC_SUCCESS

# -----------------------------------------------------------------------------



# -----------------------------------------------------------------------------

cdef extern from * nogil:
    const MPI_Comm MPI_COMM_NULL
    const MPI_Comm MPI_COMM_SELF
    const MPI_Comm MPI_COMM_WORLD

cdef extern from * nogil:
    MPI_Comm PETSC_COMM_SELF
    MPI_Comm PETSC_COMM_WORLD

# -----------------------------------------------------------------------------

from petsc4py.PETSc cimport GetComm
cdef inline MPI_Comm def_Comm(
    object comm,
    MPI_Comm defv,
) except? MPI_COMM_NULL:
    return GetComm(comm, defv)

from petsc4py.PETSc cimport GetCommDefault
cdef inline MPI_Comm SLEPC_COMM_DEFAULT() except? MPI_COMM_NULL:
    return GetCommDefault()

# -----------------------------------------------------------------------------
